"""A text animation gadget."""

import asyncio
from collections.abc import Iterable, Sequence

from ..colors import BLACK, WHITE, Color
from .animation import _check_frame_durations
from .gadget import Gadget, Point, PosHint, Size, SizeHint
from .pane import Pane
from .text import Text

__all__ = ["TextAnimation", "Point", "Size"]


class _Frame(Text):
    def on_size(self):
        pass


class TextAnimation(Gadget):
    r"""
    A text animation gadget.

    Parameters
    ----------
    frames : Iterable[str] | None, default: None
        Frames of the animation.
    frame_durations : float | int | Sequence[float| int], default: 1/12
        Time each frame is displayed. If a sequence is provided, it's length
        should be equal to number of frames.
    animation_fg_color : Color, default: WHITE
        Foreground color of animation.
    animation_bg_color : Color, default: BLACK
        Background color of animation.
    loop : bool, default: True
        Whether to restart animation after last frame.
    reverse : bool, default: False
        Whether to play animation in reverse.
    alpha : float, default: 1.0
        Transparency of gadget.
    size : Size, default: Size(10, 10)
        Size of gadget.
    pos : Point, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: False
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    frames : list[Text]
        Frames of the animation.
    frame_durations : list[int | float]
        Time each frame is displayed.
    animation_fg_color : Color
        Foreground color of animation.
    animation_bg_color : Color
        Background color of animation.
    loop : bool
        Whether to restart animation after last frame.
    reverse : bool
        Whether to play animation in reverse.
    alpha : float
        Transparency of gadget.
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : SizeHint
        Size as a proportion of parent's height and width.
    pos_hint : PosHint
        Position as a proportion of parent's height and width.
    parent: Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App
        The running app.

    Methods
    -------
    play()
        Play the animation. Returns a task.
    pause()
        Pause the animation
    stop()
        Stop the animation and reset current frame.
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(\*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    def __init__(
        self,
        *,
        frames: Iterable[str] | None = None,
        frame_durations: float | Sequence[float] = 1 / 12,
        animation_fg_color: Color = WHITE,
        animation_bg_color: Color = BLACK,
        loop: bool = True,
        reverse: bool = False,
        alpha: float = 1.0,
        size: Size = Size(10, 10),
        pos: Point = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = False,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        self.frames: Iterable[Text] = []
        if frames is not None:
            for frame in frames:
                self.frames.append(Text())
                self.frames[-1].set_text(frame)
                self.frames[-1].parent = self
        self._pane = Pane(
            size_hint={"height_hint": 1.0, "width_hint": 1.0},
            is_transparent=is_transparent,
        )
        self._frame = _Frame(is_enabled=False, is_transparent=True)
        super().__init__(
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )
        self.add_gadgets(self._pane, self._frame)
        self.frame_durations: list[float] = _check_frame_durations(
            self.frames, frame_durations
        )
        self.animation_fg_color = animation_fg_color
        self.animation_bg_color = animation_bg_color
        self.loop = loop
        self.reverse = reverse
        self.alpha = alpha
        self._i = len(self.frames) - 1 if self.reverse else 0
        self._animation_task = None

    @property
    def animation_fg_color(self) -> Color:
        """Foreground color pair of animation."""
        return self._pane.fg_color

    @animation_fg_color.setter
    def animation_fg_color(self, animation_fg_color: Color):
        self._pane.fg_color = animation_fg_color
        self._animation_fg_color = animation_fg_color
        for frame in self.frames:
            frame.canvas["fg_color"] = animation_fg_color

    @property
    def animation_bg_color(self) -> Color:
        """Foreground color pair of animation."""
        return self._pane.bg_color

    @animation_bg_color.setter
    def animation_bg_color(self, animation_bg_color: Color):
        self._pane.bg_color = animation_bg_color
        self._animation_bg_color = animation_bg_color
        for frame in self.frames:
            frame.canvas["bg_color"] = animation_bg_color

    def on_transparency(self) -> None:
        """Update gadget after transparency is enabled/disabled."""
        self._pane.is_transparent = self.is_transparent

    @property
    def alpha(self) -> bool:
        """Transparency of gadget."""
        return self._pane.alpha

    @alpha.setter
    def alpha(self, alpha: float):
        self._pane.alpha = alpha

    def on_remove(self):
        """Pause animation on remove."""
        self.pause()
        super().on_remove()

    async def _play_animation(self):
        while self.frames:
            current_frame = self.frames[self._i]
            self._frame.canvas = current_frame.canvas
            self._frame.size = current_frame.size
            self._frame.is_enabled = True
            try:
                await asyncio.sleep(self.frame_durations[self._i])
            except asyncio.CancelledError:
                self._frame.is_enabled = False
                return

            if self.reverse:
                self._i -= 1
                if self._i < 0:
                    self._i = len(self.frames) - 1

                    if not self.loop:
                        self._frame.is_enabled = False
                        return
            else:
                self._i += 1
                if self._i == len(self.frames):
                    self._i = 0

                    if not self.loop:
                        self._frame.is_enabled = False
                        return

    def play(self) -> asyncio.Task:
        """
        Play animation.

        Returns
        -------
        asyncio.Task
            The task that plays the animation.
        """
        self.pause()

        if self._i == 0 and self.reverse:
            self._i = len(self.frames) - 1
        elif self._i == len(self.frames) - 1 and not self.reverse:
            self._i = 0

        self._animation_task = asyncio.create_task(self._play_animation())
        return self._animation_task

    def pause(self):
        """Pause animation."""
        if self._animation_task is not None:
            self._animation_task.cancel()

    def stop(self):
        """Stop the animation and reset current frame."""
        self.pause()
        self._i = len(self.frames) - 1 if self.reverse else 0



"""Platform specific VT100 terminals."""

import platform
import sys
from collections.abc import Callable
from contextlib import contextmanager
from typing import ContextManager

from .events import Event
from .vt100_terminal import Vt100Terminal

__all__ = ["Vt100Terminal", "get_platform_terminal", "app_mode"]


def get_platform_terminal() -> Vt100Terminal:
    """
    Return a platform-specific terminal.

    Returns
    -------
    Vt100Terminal
        A platform-specific VT100 terminal.

    Raises
    ------
    RuntimeError
        If terminal isn't interactive or terminal doesn't support VT100 sequences.
    """
    if not sys.stdin.isatty():
        raise RuntimeError("Terminal is non-interactive.")

    if platform.system() == "Windows":
        from .windows_terminal import WindowsTerminal, is_vt100_enabled

        if not is_vt100_enabled():
            raise RuntimeError("Terminal doesn't support VT100 sequences.")

        return WindowsTerminal()
    else:
        from .linux_terminal import LinuxTerminal

        return LinuxTerminal()


@contextmanager
def app_mode(
    terminal: Vt100Terminal, event_handler: Callable[[list[Event]], None]
) -> ContextManager[None]:
    """
    Put terminal into app mode and dispatch input events.

    Parameters
    ----------
    terminal : Vt100Terminal
        Terminal to put in app mode.
    event_handler : Callable[[list[Event]], None]
        A callable that handles terminal input events.
    """
    try:
        terminal.raw_mode()
        terminal.attach(event_handler)
        terminal.enable_mouse_support()
        terminal.enable_bracketed_paste()
        terminal.enable_reporting_focus()
        terminal.hide_cursor()
        terminal.flush()
        yield
    finally:
        if terminal.in_alternate_screen:
            terminal.exit_alternate_screen()
        else:
            terminal.erase_in_display()
        terminal.reset_attributes()
        terminal.show_cursor()
        terminal.disable_reporting_focus()
        terminal.disable_bracketed_paste()
        terminal.disable_mouse_support()
        terminal.flush()
        terminal.unattach()
        terminal.restore_console()



"""Color data structures."""

import re
from typing import NamedTuple, Self, TypedDict

__all__ = ["AColor", "Color", "ColorTheme"]


def validate_hexcode(hexcode: str) -> bool:
    return bool(re.match(r"^#?[0-9a-fA-F]{6}$", hexcode))


def validate_ahexcode(ahexcode: str) -> bool:
    return bool(re.match(r"^#?[0-9a-fA-F]{8}$", ahexcode))


def _to_hex(channel):
    return hex(channel)[2:].zfill(2)


class Color(NamedTuple):
    """
    A 24-bit color.

    Parameters
    ----------
    red : int
        The red component.
    green : int
        The green component.
    blue : int
        The blue component.

    Attributes
    ----------
    red : int
        The red component.
    green : int
        The green component.
    blue : int
        The blue component.

    Methods
    -------
    from_hex(hexcode)
        Create a :class:`Color` from a hex code.
    to_hex()
        Return color's hexcode.
    """

    red: int
    green: int
    blue: int

    @classmethod
    def from_hex(cls, hexcode: str) -> Self:
        """
        Create a :class:`Color` from a hex code.

        Parameters
        ----------
        hexcode : str
            A color hex code.

        Returns
        -------
        Color
            A new color.
        """
        if not validate_hexcode(hexcode):
            raise ValueError(f"{hexcode!r} is not a valid hex code.")

        digits = hexcode.removeprefix("#")
        return cls(int(digits[:2], 16), int(digits[2:4], 16), int(digits[4:], 16))

    def to_hex(self) -> str:
        """
        Return color's hexcode.

        Returns
        -------
        str
            The hexcode of the color.
        """
        hexcode = "".join(_to_hex(channel) for channel in self)
        return f"#{hexcode}"


class AColor(NamedTuple):
    """
    A 32-bit color with an alpha channel.

    Parameters
    ----------
    red : int
        The red component.
    green : int
        The green component.
    blue : int
        The blue component.
    alpha : int, default: 255
        The alpha component.

    Attributes
    ----------
    red : int
        The red component.
    green : int
        The green component.
    blue : int
        The blue component.
    alpha : int
        The alpha component.

    Methods
    -------
    from_hex(hexcode)
        Create an :class:`AColor` from a hex code.
    to_hex()
        Return color's hexcode.
    """

    red: int
    green: int
    blue: int
    alpha: int = 255

    @classmethod
    def from_hex(cls, hexcode: str) -> Self:
        """
        Create an :class:`AColor` from a hex code.

        Parameters
        ----------
        hexcode : str
            A color hex code.

        Returns
        -------
        AColor
            A new color with alpha.
        """
        if not validate_ahexcode(hexcode) and not validate_hexcode(hexcode):
            raise ValueError(f"{hexcode!r} is not a valid hex code")

        digits = hexcode.removeprefix("#")
        return cls(
            int(digits[:2], 16),
            int(digits[2:4], 16),
            int(digits[4:6], 16),
            int(digits[6:] or "ff", 16),
        )

    def to_hex(self) -> str:
        """
        Return color's hexcode.

        Returns
        -------
        str
            The hexcode of the color.
        """
        hexcode = "".join(_to_hex(channel) for channel in self)
        return f"#{hexcode}"


class ColorPair(TypedDict):
    """
    A foreground and background hexcode.

    Attributes
    ----------
    fg : str
        The foreground hex code.
    bg : str
        The background hex code.
    """

    fg: str
    """The foreground hex code."""
    bg: str
    """The background hex code."""


class ColorTheme(TypedDict, total=False):
    """
    Colors for themable gadgets.

    Missing colors will use the default color theme.

    Attributes
    ----------
    primary : ColorPair
        The primary color pair.
    text_pad_line_highlight : ColorPair
        Text pad line highlight color pair.
    text_pad_selection_highlight : ColorPair
        Text pad selection color pair.
    textbox_primary : ColorPair
        Text pad primary color pair.
    textbox_selection_highlight : ColorPair
        Textbox selection color pair.
    textbox_placeholder : ColorPair
        Textbox placeholder color pair.
    button_normal : ColorPair
        Button normal color pair.
    button_hover : ColorPair
        Button hover color pair.
    button_press : ColorPair
        Button press color pair.
    button_disallowed : ColorPair
        Button disallowed color pair.
    menu_item_hover : ColorPair
        Menu item hover color pair.
    menu_item_selected : ColorPair
        Menu item selected color pair.
    menu_item_disallowed : ColorPair
        Menu item disallowed color pair.
    titlebar_normal : ColorPair
        Titlebar normal color pair.
    titlebar_inactive : ColorPair
        Titlebar inactive color pair.
    data_table_sort_indicator : ColorPair
        Data table sort indicator color pair.
    data_table_hover : ColorPair
        Data table hover color pair.
    data_table_stripe : ColorPair
        Data table stripe color pair.
    data_table_stripe_hover : ColorPair
        Data table stripe hover color pair.
    data_table_selected : ColorPair
        Data table selected color pair.
    data_table_selected_hover : ColorPair
        Data table selected hover color pair.
    progress_bar : ColorPair
        Progress bar color pair.
    markdown_link : ColorPair
        Markdown link color pair.
    markdown_link_hover : ColorPair
        Markdown link hover color pair.
    markdown_inline_code : ColorPair
        Markdown inline code color pair.
    markdown_quote : ColorPair
        Markdown quote color pair.
    markdown_title : ColorPair
        Markdown title color pair.
    markdown_image : ColorPair
        Markdown image color pair.
    markdown_block_code_background : str
        Markdown block code background color hexcode.
    markdown_quote_block_code_background : str
        Markdown quote block code background color hexcode.
    markdown_header_background : str
        Markdown header background color hexcode.
    scroll_view_scrollbar : str
        Scroll view scrollbar color hexcode.
    scroll_view_indicator_normal : str
        Scroll view indicator normal color hexcode.
    scroll_view_indicator_hover : str
        Scroll view indicator hover color hexcode.
    scroll_view_indicator_press : str
        Scroll view indicator press color hexcode.
    window_border_normal : str
        Window border normal color hexcode.
    window_border_inactive : str
        Window border inactive color hexcode.
    """

    primary: ColorPair
    """The primary color pair."""
    text_pad_line_highlight: ColorPair
    """Text pad line highlight color pair."""
    text_pad_selection_highlight: ColorPair
    """Text pad selection color pair."""
    textbox_primary: ColorPair
    """Text pad primary color pair."""
    textbox_selection_highlight: ColorPair
    """Textbox selection color pair."""
    textbox_placeholder: ColorPair
    """Textbox placeholder color pair."""
    button_normal: ColorPair
    """Button normal color pair."""
    button_hover: ColorPair
    """Button hover color pair."""
    button_press: ColorPair
    """Button press color pair."""
    button_disallowed: ColorPair
    """Button disallowed color pair."""
    menu_item_hover: ColorPair
    """Menu item hover color pair."""
    menu_item_selected: ColorPair
    """Menu item selected color pair."""
    menu_item_disallowed: ColorPair
    """Menu item disallowed color pair."""
    titlebar_normal: ColorPair
    """Titlebar normal color pair."""
    titlebar_inactive: ColorPair
    """Titlebar inactive color pair."""
    data_table_sort_indicator: ColorPair
    """Data table sort indicator color pair."""
    data_table_hover: ColorPair
    """Data table hover color pair."""
    data_table_stripe: ColorPair
    """Data table stripe color pair."""
    data_table_stripe_hover: ColorPair
    """Data table stripe hover color pair."""
    data_table_selected: ColorPair
    """Data table selected color pair."""
    data_table_selected_hover: ColorPair
    """Data table selected hover color pair."""
    progress_bar: ColorPair
    """Progress bar color pair."""
    markdown_link: ColorPair
    """Markdown link color pair."""
    markdown_link_hover: ColorPair
    """Markdown link hover color pair."""
    markdown_inline_code: ColorPair
    """Markdown inline code color pair."""
    markdown_quote: ColorPair
    """Markdown quote color pair."""
    markdown_title: ColorPair
    """Markdown title color pair."""
    markdown_image: ColorPair
    """Markdown image color pair."""
    markdown_block_code_background: str
    """Markdown block code background color hexcode."""
    markdown_quote_block_code_background: str
    """Markdown quote block code background color hexcode."""
    markdown_header_background: str
    """Markdown header background color hexcode."""
    scroll_view_scrollbar: str
    """Scroll view scrollbar color hexcode."""
    scroll_view_indicator_normal: str
    """Scroll view indicator normal color hexcode."""
    scroll_view_indicator_hover: str
    """Scroll view indicator hover color hexcode."""
    scroll_view_indicator_press: str
    """Scroll view indicator press color hexcode."""
    window_border_normal: str
    """Window border normal color hexcode."""
    window_border_inactive: str
    """Window border inactive color hexcode."""



"""Tools for text."""

from bisect import bisect
from functools import lru_cache
from operator import itemgetter

import numpy as np
from numpy.typing import NDArray

from ._batgrl_markdown import find_md_tokens
from ._char_widths import CHAR_WIDTHS
from .colors import BLACK, WHITE, Color
from .geometry import Size

__all__ = [
    "Cell",
    "add_text",
    "binary_to_box",
    "binary_to_braille",
    "new_cell",
    "char_width",
    "coerce_cell",
    "is_word_char",
    "smooth_horizontal_bar",
    "smooth_vertical_bar",
    "str_width",
]

VERTICAL_BLOCKS = " ▁▂▃▄▅▆▇█"
HORIZONTAL_BLOCKS = " ▏▎▍▌▋▊▉█"
_BRAILLE_ENUM = np.array([[1, 8], [2, 16], [4, 32], [64, 128]])
_BOX_ENUM = np.array([[1, 4], [2, 8]])

_vectorized_chr = np.vectorize(chr)
"""Vectorized `chr`."""

_vectorized_box_map = np.vectorize(" ▘▖▌▝▀▞▛▗▚▄▙▐▜▟█".__getitem__)
"""Vectorized box enum to box char."""


@lru_cache(maxsize=1024)
def char_width(char: str) -> int:
    """
    Return the column width of a character.

    Parameters
    ----------
    char : str
        A unicode character.

    Returns
    -------
    int
        The character column width.
    """
    if char == "":
        return 0

    char_ord = ord(char)
    i = bisect(CHAR_WIDTHS, char_ord, key=itemgetter(0))
    if i == 0:
        return 1

    _, high, width = CHAR_WIDTHS[i - 1]
    if char_ord <= high:
        return width

    return 1


@lru_cache(maxsize=256)
def str_width(chars: str) -> int:
    """
    Return the total column width of a string.

    Parameters
    ----------
    chars : str
        A string.

    Returns
    -------
    int
        The total column width of the string.
    """
    return sum(map(char_width, chars))


def is_word_char(char: str) -> bool:
    """
    Whether `char` is a word character.

    A character is a word character if it is alphanumeric or an underscore.

    Parameters
    ----------
    char : str
        The char to test.

    Returns
    -------
    bool
        Whether the char is a word character.
    """
    return char.isalnum() or char == "_"


Cell = np.dtype(
    [
        ("char", "U1"),
        ("bold", "?"),
        ("italic", "?"),
        ("underline", "?"),
        ("strikethrough", "?"),
        ("overline", "?"),
        ("reverse", "?"),
        ("fg_color", "u1", (3,)),
        ("bg_color", "u1", (3,)),
    ]
)
"""A structured array type that represents a single cell in a terminal."""


@lru_cache
def cell_sans(*names: str) -> list[str]:
    r"""
    Return all fields of ``Cell`` not in names.

    Parameters
    ----------
    \*names : str
        Excluded fields of ``Cell``.

    Returns
    -------
    list[str]
        All fields of ``Cell`` not in names.
    """
    return [name for name in Cell.names if name not in names]


def new_cell(
    char: str = " ",
    bold: bool = False,
    italic: bool = False,
    underline: bool = False,
    strikethrough: bool = False,
    overline: bool = False,
    reverse: bool = False,
    fg_color: Color = WHITE,
    bg_color: Color = BLACK,
) -> NDArray[Cell]:
    """
    Create a ``Cell`` scalar.

    A Cell is a structured array type that represents a single cell in a terminal.

    Parameters
    ----------
    char : str, default: " "
        The cell's character.
    bold : bool, default: False
        Whether cell is bold.
    italic : bool, default: False
        Whether cell is italic.
    underline : bool, default: False
        Whether cell is underlined.
    strikethrough : bool, default: False
        Whether cell is strikethrough.
    overline : bool, default: False
        Whether cell is overlined.
    fg_color : Color, default: WHITE
        Foreground color of cell.
    bg_color : Color, default: BLACK
        Background color of cell.

    Returns
    -------
    NDArray[Cell]
        A ``Cell`` scalar.
    """
    return np.array(
        (
            char,
            bold,
            italic,
            underline,
            strikethrough,
            overline,
            reverse,
            fg_color,
            bg_color,
        ),
        dtype=Cell,
    )


def coerce_cell(char: NDArray[Cell] | str, default: NDArray[Cell]) -> NDArray[Cell]:
    """
    Try to coerce a string or ``Cell`` scalar into a half-width ``Cell`` scalar.

    Parameters
    ----------
    char : NDArray[Cell] | str
        The character to coerce.
    default : NDArray[Cell] | None, default: None
        The fallback character (or None) if character can't be coerced.

    Returns
    -------
    NDArray[Cell] | None
        The coerced Cell or None if character can't be coerced.
    """
    if isinstance(char, str) and len(char) > 0 and char_width(char[0]) == 1:
        return new_cell(char=char[0])
    if (
        isinstance(char, np.ndarray)
        and char.dtype == Cell
        and char.shape == ()
        and char_width(char["char"].item()) == 1
    ):
        return char
    return default


def _parse_batgrl_md(text: str) -> tuple[Size, list[list[NDArray[Cell]]]]:
    """
    Parse batgrl markdown and return the minimum canvas size to fit text and
    a list of lines of styled characters.

    #### Syntax for batgrl markdown
    - italic: `*this is italic text*`
    - bold: `**this is bold text**`
    - strikethrough: `~~this is strikethrough text~~`
    - underlined: `__this is underlined text__`
    - overlined: `^^this is overlined text^^`

    Parameters
    ----------
    text : str
        The text to parse.

    Returns
    -------
    tuple[Size, list[list[Cell]]]
        Minimum canvas size to fit text and a list of lines of styled characters.
    """
    NO_CHAR = new_cell(char="")
    matches, escapes = find_md_tokens(text)
    cells = [new_cell(char=char)[cell_sans("fg_color", "bg_color")] for char in text]
    for before, start, end, after, style in matches:
        cells[start - before : start] = [NO_CHAR] * before
        cells[end : end + after] = [NO_CHAR] * after
        for i in range(start, end):
            cells[i][style] = True

    for i in escapes:
        cells[i] = NO_CHAR

    lines = []
    line = []
    line_width = max_width = 0

    for cell in cells:
        char = cell["char"].item()
        if char == "":
            continue

        if char == "\n":
            lines.append(line)
            line = []
            if line_width > max_width:
                max_width = line_width
            line_width = 0
        else:
            width = char_width(char)
            line_width += width
            if width > 0:
                line.append(cell)

    lines.append(line)
    if line_width > max_width:
        max_width = line_width

    return Size(len(lines), max_width), lines


def _text_to_cells(text: str) -> tuple[Size, list[list[NDArray[Cell]]]]:
    """
    Convert some text to a list of lists of Cells and the minimum canvas size to fit
    them.

    Parameters
    ----------
    text : str
        The text to convert.

    Returns
    -------
    tuple[Size, list[list[Cell]]]
        Minimum canvas size to fit text and a list of lists of Cells.
    """
    lines = [
        [new_cell(char=char)[cell_sans("fg_color", "bg_color")] for char in line]
        for line in text.split("\n")
    ]
    max_width = 0
    for line in lines:
        line_width = 0
        for char in line:
            width = char_width(char["char"].item())
            line_width += width
        if line_width > max_width:
            max_width = line_width
    return Size(len(lines), max_width), lines


def _write_lines_to_canvas(lines, canvas, fg_color, bg_color):
    """Write a list of lists of Cells to a canvas array."""
    _, columns = canvas.shape
    for cells, canvas_line, fg, bg in zip(
        lines,
        canvas[cell_sans("fg_color", "bg_color")],
        canvas["fg_color"],
        canvas["bg_color"],
    ):
        i = 0
        for cell in cells:
            if i >= columns:
                break

            width = char_width(cell["char"].item())
            if width == 0:
                continue

            canvas_line[i] = cell
            if fg_color is not None:
                fg[i] = fg_color
            if bg_color is not None:
                bg[i] = bg_color

            if width == 2 and i + 1 < columns:
                empty_cell = cell.copy()
                empty_cell["char"] = ""
                canvas_line[i + 1] = empty_cell
                if fg_color is not None:
                    fg[i + 1] = fg_color
                if bg_color is not None:
                    bg[i + 1] = bg_color

            i += width


def add_text(
    canvas: NDArray[Cell],
    text: str,
    *,
    fg_color: Color | None = None,
    bg_color: Color | None = None,
    markdown: bool = False,
    truncate_text: bool = False,
):
    """
    Add multiple lines of text to a view of a canvas.

    If `markdown` is true, text can be styled using batgrl markdown. The syntax is:
    - italic: `*this is italic text*`
    - bold: `**this is bold text**`
    - strikethrough: `~~this is strikethrough text~~`
    - underlined: `__this is underlined text__`
    - overlined: `^^this is overlined text^^`

    Text is added starting at first index in canvas. Every new line is added on a new
    row.

    Parameters
    ----------
    canvas : NDArray[Cell]
        A 1- or 2-dimensional view of a `Text` canvas.
    text : str
        Text to add to canvas.
    fg_color : Color | None, default: None
        Foreground color of text. If not given, current canvas foreground color is used.
    bg_color : Color | None, default: None
        Background color of text. If not given, current canvas background color is used.
    markdown : bool, default: False
        Whether to parse text for batgrl markdown.
    truncate_text : bool, default: False
        For text that doesn't fit on canvas, truncate text if true else raise an
        `IndexError`.
    """
    size, lines = _parse_batgrl_md(text) if markdown else _text_to_cells(text)
    if canvas.ndim == 1:  # Pre-pend an axis if canvas is one-dimensional.
        canvas = canvas[None]
    rows, columns = canvas.shape
    if not truncate_text and (size.height > rows or size.width > columns):
        raise IndexError("Text does not fit in canvas.")
    _write_lines_to_canvas(lines, canvas, fg_color, bg_color)


def _smooth_bar(
    blocks: str,
    max_length: int,
    proportion: float,
    offset: float,
):
    """Create a smooth bar with given blocks."""
    if offset >= 1 or offset < 0:
        raise ValueError(
            f"Offset should greater than or equal to 0 and less than 1, but {offset} "
            "was given."
        )

    block_indices = len(blocks) - 1
    fill, partial = divmod(proportion * max_length, 1)
    fill = int(fill)

    if offset == 0.0:
        if fill == max_length:
            return (blocks[-1],) * fill

        index_partial = round(partial * block_indices)
        partial_block = blocks[index_partial]
        return (*(blocks[-1],) * fill, partial_block)

    partial += offset
    if partial > 1:
        partial -= 1
    else:
        fill -= 1

    index_offset = round(offset * block_indices)
    index_partial = round(partial * block_indices)
    offset_block = blocks[index_offset]
    partial_block = blocks[index_partial]
    return (offset_block, *(blocks[-1],) * fill, partial_block)


def smooth_vertical_bar(
    max_height: int,
    proportion: float,
    offset: float = 0.0,
    reversed: bool = False,
) -> tuple[str, ...]:
    """
    Create a vertical bar that's some proportion of max_height at an offset.

    Offset bars will return a minimum of 2 characters and the first character of the bar
    should have it's colors reversed (or, if bar is reversed, all colors should be
    reversed except first character).

    Parameters
    ----------
    max_height : int
        The height of the bar if proportion was 1.0.
    proportion : float
        The height of the bar as a proportion of the max_height.
    offset : float, default: 0.0
        Offset the bar vertically by some non-negative amount.
    reversed : bool, default: False
        Reversed vertical bar is drawn top-to-bottom and offset downwards.

    Returns
    -------
    tuple[str, ...]
        The bar as a tuple of characaters.
    """
    blocks = VERTICAL_BLOCKS[::-1] if reversed else VERTICAL_BLOCKS
    return _smooth_bar(blocks, max_height, proportion, offset)


def smooth_horizontal_bar(
    max_width: int, proportion: float, offset: float = 0.0
) -> tuple[str, ...]:
    """
    Create a horizontal bar that's some proportion of max_width at an offset.

    Offset bars will return a minimum of 2 characters and the first character of the bar
    should have it's colors reversed.

    Parameters
    ----------
    max_width : int
        The width of the bar if the proportion was 1.0.
    proportion : float
        The width of the bar as a proportion of the max_width.
    offset : float, default: 0.0
        Offset the bar horizontally by some non-negative amont.

    Returns
    -------
    tuple[str, ...]
        The bar as a tuple of characters.
    """
    return _smooth_bar(HORIZONTAL_BLOCKS, max_width, proportion, offset)


def binary_to_braille(array_4x2: NDArray[np.bool_]) -> NDArray[np.dtype("<U1")]:
    r"""
    Convert a (h, w, 4, 2)-shaped boolean array into a (h, w) array of braille unicode
    characters.

    Parameters
    ----------
    array_4x2 : NDArray[np.bool\_]
        A (h, w, 4, 2)-shaped boolean numpy array.

    Returns
    -------
    NDArray[np.dtype("<U1")]
        A numpy array of braille unicode characters.
    """
    return _vectorized_chr(
        np.sum(array_4x2 * _BRAILLE_ENUM, axis=(2, 3), initial=0x2800)
    )


def binary_to_box(array_2x2: NDArray[np.bool_ | np.uint]) -> NDArray[np.dtype("<U1")]:
    r"""
    Convert a (h, w, 2, 2)-shaped boolean array into a (h, w) array of box unicode
    characters.

    Parameters
    ----------
    array_2x2 : NDArray[np.bool\_]
        A (h, w, 2, 2)-shaped boolean numpy array.

    Returns
    -------
    NDArray[np.dtype("<U1")]
        A numpy array of box unicode characters.
    """
    return _vectorized_box_map(np.sum(array_2x2 * _BOX_ENUM, axis=(2, 3)))



"""A graphic gadget."""

from pathlib import Path

import cv2
import numpy as np
from numpy.typing import NDArray

from ..colors import TRANSPARENT, AColor
from ..geometry import rect_slice
from ..text_tools import cell_sans
from ..texture_tools import Interpolation, _composite, resize_texture
from .gadget import Cell, Gadget, Point, PosHint, Size, SizeHint, bindable, clamp

__all__ = ["Graphics", "Interpolation", "Point", "Size"]


class Graphics(Gadget):
    r"""
    A graphic gadget. Displays arbitrary RGBA textures.

    Graphic gadgets are gadgets that are rendered entirely with the upper half block
    character, "▀". Graphic gadgets' color information is stored in a uint8 RGBA array,
    :attr:`texture`. Note that the height of :attr:`texture` is twice the height of the
    gadget.

    Parameters
    ----------
    default_color : AColor, default: AColor(0, 0, 0, 0)
        Default texture color.
    alpha : float, default: 1.0
        Transparency of gadget.
    interpolation : Interpolation, default: "linear"
        Interpolation used when gadget is resized.
    size : Size, default: Size(10, 10)
        Size of gadget.
    pos : Point, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: True
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    texture : NDArray[np.uint8]
        uint8 RGBA color array.
    default_color : AColor
        Default texture color.
    alpha : float
        Transparency of gadget.
    interpolation : Interpolation
        Interpolation used when gadget is resized.
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : SizeHint
        Size as a proportion of parent's height and width.
    pos_hint : PosHint
        Position as a proportion of parent's height and width.
    parent: Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App
        The running app.

    Methods
    -------
    to_png(path)
        Write :attr:`texture` to provided path as a `png` image.
    clear()
        Fill texture with default color.
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(\*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    def __init__(
        self,
        *,
        is_transparent: bool = True,
        default_color: AColor = TRANSPARENT,
        alpha: float = 1.0,
        interpolation: Interpolation = "linear",
        size: Size = Size(10, 10),
        pos: Point = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        super().__init__(
            is_transparent=is_transparent,
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )

        self.default_color = default_color
        self.alpha = alpha
        self.interpolation = interpolation

        h, w = self.size
        self.texture = np.full((2 * h, w, 4), default_color, dtype=np.uint8)

    @property
    def alpha(self) -> float:
        """Transparency of gadget."""
        return self._alpha

    @alpha.setter
    @bindable
    def alpha(self, alpha: float):
        self._alpha = clamp(float(alpha), 0.0, 1.0)

    @property
    def interpolation(self) -> Interpolation:
        """Interpolation used when gadget is resized."""
        return self._interpolation

    @interpolation.setter
    def interpolation(self, interpolation: Interpolation):
        if interpolation not in Interpolation.__args__:
            raise TypeError(f"{interpolation} is not a valid interpolation type.")
        self._interpolation = interpolation

    def on_size(self):
        """Resize texture array."""
        h, w = self.size
        self.texture = resize_texture(self.texture, (2 * h, w), self.interpolation)

    def _render(self, canvas: NDArray[Cell]):
        """Render visible region of gadget."""
        texture = self.texture
        chars = canvas["char"]
        styles = canvas[cell_sans("char", "fg_color", "bg_color")]
        foreground = canvas["fg_color"]
        background = canvas["bg_color"]
        root_pos = self.root._pos
        abs_pos = self.absolute_pos
        alpha = self.alpha
        for pos, (h, w) in self._region.rects():
            dst = rect_slice(pos - root_pos, (h, w))
            src_top, src_left = pos - abs_pos
            src_bottom, src_right = src_top + h, src_left + w
            fg_rect = foreground[dst]
            bg_rect = background[dst]
            even_rows = texture[2 * src_top : 2 * src_bottom : 2, src_left:src_right]
            odd_rows = texture[2 * src_top + 1 : 2 * src_bottom : 2, src_left:src_right]

            if self.is_transparent:
                mask = chars[dst] != "▀"
                fg_rect[mask] = bg_rect[mask]
                _composite(fg_rect, even_rows[..., :3], even_rows[..., 3, None], alpha)
                _composite(bg_rect, odd_rows[..., :3], odd_rows[..., 3, None], alpha)
            else:
                fg_rect[:] = even_rows[..., :3]
                bg_rect[:] = odd_rows[..., :3]

            chars[dst] = "▀"
            styles[dst] = False

    def to_png(self, path: Path):
        """Write :attr:`texture` to provided path as a `png` image."""
        BGRA = cv2.cvtColor(self.texture, cv2.COLOR_RGBA2BGRA)
        cv2.imwrite(str(path.absolute()), BGRA)

    def clear(self):
        """Fill texture with default color."""
        self.texture[:] = self.default_color



"""Tools for graphics."""

from pathlib import Path
from typing import Literal

import cv2
import numpy as np
from numpy.typing import NDArray

from .geometry import Point, Region, Size, rect_slice

__all__ = ["Interpolation", "read_texture", "resize_texture", "composite"]

Interpolation = Literal["nearest", "linear", "cubic", "area", "lanczos"]
"""Interpolation methods for resizing graphic gadgets."""

_INTERPOLATION_TO_CV_ENUM = {
    "linear": cv2.INTER_LINEAR,
    "cubic": cv2.INTER_CUBIC,
    "area": cv2.INTER_AREA,
    "lanczos": cv2.INTER_LANCZOS4,
    "nearest": cv2.INTER_NEAREST,
}


def read_texture(path: Path) -> NDArray[np.uint8]:
    """
    Return a uint8 RGBA numpy array from a path to an image.

    Parameters
    ----------
    path : Path
        Path to image.

    Returns
    -------
    NDArray[np.uint8]
        An uint8 RGBA array of the image.
    """
    image = cv2.imread(str(path.absolute()), cv2.IMREAD_UNCHANGED)

    if image.dtype == np.dtype(np.uint16):
        image = (image // 257).astype(np.uint8)
    elif image.dtype == np.dtype(np.float32):
        image = (image * 255).astype(np.uint8)

    # Add an alpha channel if there isn't one.
    h, w, c = image.shape
    if c == 3:
        default_alpha_channel = np.full((h, w, 1), 255, dtype=np.uint8)
        image = np.dstack((image, default_alpha_channel))

    return cv2.cvtColor(image, cv2.COLOR_BGRA2RGBA)


def resize_texture(
    texture: NDArray[np.uint8], size: Size, interpolation: Interpolation = "linear"
) -> NDArray[np.uint8]:
    """
    Resize texture.

    Parameters
    ----------
    texture : NDArray[np.uint8]
        An RGBA texture to resize.
    size : Size
        The new size of the texture.
    interpolation : Interpolation, default: "linear"
        Interpolation used when resizing texture.

    Returns
    -------
    NDArray[np.uint8]
        A new uint8 RGBA array.
    """
    old_h, old_w = texture.shape[:2]
    h, w = size
    if old_h == 0 or old_w == 0 or h == 0 or w == 0:
        return np.zeros((h, w, 4), np.uint8)
    return cv2.resize(
        texture, (w, h), interpolation=_INTERPOLATION_TO_CV_ENUM[interpolation]
    )


def _composite(
    dest: NDArray[np.uint8],
    rgb: tuple[int, int, int] | NDArray[np.uint8],
    a: int | NDArray[np.uint8],
    alpha: float = 1.0,
) -> None:
    """
    Composite an rgb texture or color onto ``dest``.

    This is an internal function used in various ``_render()`` methods.
    """
    buffer = np.subtract(rgb, dest, dtype=float)
    buffer *= a
    buffer *= alpha / 255
    np.add(buffer, dest, out=dest, casting="unsafe")


def composite(
    source: NDArray[np.uint8],
    dest: NDArray[np.uint8],
    pos: Point = Point(0, 0),
    mask_mode: bool = False,
) -> None:
    """
    Composite source texture onto destination texture at given position.

    If `mask_mode` is true, source alpha values less than 255 are ignored.

    Parameters
    ----------
    source : NDArray[np.uint8]
        The texture to composite.
    dest : NDArray[np.uint8]
        The texture on which the source is painted.
    pos : Point, default: Point(0, 0)
        Position of the source on the destination.
    mask_mode : bool, default: False
        Whether to ignore alpha values less than 255.
    """
    sh, sw, _ = source.shape
    dh, dw, _ = dest.shape

    dest_reg = Region.from_rect((0, 0), (dh, dw))
    source_reg = Region.from_rect(pos, (sh, sw))

    if intersection := source_reg & dest_reg:
        rpos, size = next(intersection.rects())
        dest_tex = dest[rect_slice(rpos, size)]
        source_tex = source[rect_slice(rpos - pos, size)]
        source_alpha = source_tex[..., 3]

        if mask_mode:
            mask = source_alpha == 255
            dest_tex[mask] = source_tex[mask]
        else:
            _composite(dest_tex, source_tex, source_alpha[..., None])



"""Root gadget."""

from __future__ import annotations

from threading import RLock
from typing import TYPE_CHECKING, Literal, Self

import numpy as np
from numpy.typing import NDArray

if TYPE_CHECKING:
    from ..app import App

from ..text_tools import Cell, new_cell
from .gadget import Gadget, Point, Region, Size, _GadgetList


class _Root(Gadget):
    """
    Root gadget of the gadget tree.

    Instantiated only by :class:`batgrl.app.App`.
    """

    def __init__(self, app: App, size: Size):
        self.children = _GadgetList()

        self._render_lock = RLock()
        """Lock held during rendering to prevent errors related to invalid geometry."""
        self._app = app
        """The running app."""
        self._cell = new_cell()
        """Default cell of root canvas."""
        self._all_regions_valid = False
        """Whether all regions in gadget tree are valid."""
        self._region_valid = False

        self._resized: bool
        """Whether terminal has resized since last render."""
        self._last_canvas: NDArray[Cell]
        """Previous rendering of gadget tree."""
        self.canvas: NDArray[Cell]
        """Current rendering of gadget tree."""

        self._pos = Point(0, 0)
        self._size = size
        """Size of root gadget."""
        self.on_size()

    def __repr__(self):
        return f"_Root(size={self.size}, pos={self.pos})"

    def on_size(self):
        """Remake buffers and set ``_resized`` flag on resize."""
        self._resized = True
        self._last_canvas = np.full(self._size, self._cell)
        self.canvas = self._last_canvas.copy()

    @property
    def is_transparent(self) -> Literal[False]:
        return False

    @property
    def is_visible(self) -> Literal[True]:
        return True

    @property
    def is_enabled(self) -> Literal[True]:
        return True

    @property
    def parent(self) -> Literal[None]:
        return None

    @property
    def root(self) -> Self:
        return self

    @property
    def app(self) -> App:
        """The running app."""
        return self._app

    def _set_regions(self) -> None:
        """Recompute valid regions for all gadgets with invalid regions."""
        if not self._region_valid:
            self._clipping_region = Region.from_rect(self.absolute_pos, self.size)
            self._region = self._clipping_region

        for child in self.walk():
            if not child._region_valid:
                child._clipping_region = (
                    child.parent._clipping_region
                    & Region.from_rect(child.absolute_pos, child.size)
                    if child._is_enabled and child._is_visible
                    else Region()
                )

        skip_valid_regions = True
        for child in self.walk_reverse():
            if skip_valid_regions and child._region_valid:
                continue

            if child._region_valid and child._root_region_before == self._region:
                skip_valid_regions = True
                continue

            child._root_region_before = self._region
            child._region = self._region & child._clipping_region
            if not child._is_transparent:
                self._region -= child._region

            child._region_valid = True
            skip_valid_regions = False

    def _render(self):
        """Render gadget tree into :attr:``canvas``."""
        with self._render_lock:
            if not self._all_regions_valid:
                self._set_regions()

            self.canvas, self._last_canvas = self._last_canvas, self.canvas
            self.canvas[:] = self._cell

            for child in self.walk():
                if child._is_enabled and child._is_visible:
                    child._render(self.canvas)

            self._all_regions_valid = True
            self._resized = False



"""A textbox gadget for single-line editable text."""

from __future__ import annotations

from collections.abc import Callable
from dataclasses import astuple

from numpy.typing import NDArray

from ..geometry import rect_slice
from ..terminal.events import KeyEvent, MouseButton, MouseEvent, PasteEvent
from ..text_tools import is_word_char, str_width
from ._cursor import Cursor
from .behaviors.focusable import Focusable
from .behaviors.grabbable import Grabbable
from .behaviors.themable import Themable
from .gadget import Cell, Gadget, Point, PosHint, Region, Size, SizeHint
from .text import Text

__all__ = ["Textbox", "Point", "Size"]


class _Box(Text):
    def _render(self, canvas):
        super()._render(canvas)
        textbox: Textbox = self.parent
        if textbox.hide_input:
            hider_rect = Region.from_rect(self.absolute_pos, (1, textbox._line_length))
            hider_region = self._region & hider_rect
            for pos, size in hider_region.rects():
                canvas["char"][rect_slice(pos, size)] = textbox.hide_char


class _Cursor(Cursor):
    def _render(self, canvas):
        textbox: Textbox = self.parent.parent
        placeholder = textbox._placeholder_gadget
        root_pos = self.root._pos
        abs_pos = self.parent.absolute_pos
        for pos, size in self._region.rects():
            dst = rect_slice(pos - root_pos, size)
            src = rect_slice(pos - abs_pos, size)
            canvas[dst]["fg_color"] = self.fg_color
            canvas[dst]["bg_color"] = self.bg_color
            if pos.x > textbox._line_length:
                continue
            if placeholder.is_enabled:
                canvas[dst]["char"] = placeholder.canvas[src]["char"]
            elif textbox.hide_input:
                canvas[dst]["char"] = textbox.hide_char
            else:
                canvas[dst]["char"] = textbox._box.canvas[src]["char"]


class Textbox(Themable, Focusable, Grabbable, Gadget):
    r"""
    A textbox gadget for single-line editable text.

    Supports pasting, mouse selection, and cursor navigation.

    Parameters
    ----------
    enter_callback : Callable[[Textbox], None] | None, default: None
        Called when textbox has focus and `enter` is pressed.
    placeholder : str, default: ""
        Placeholder text for textbox.
    hide_input : bool, default: False
        Whether input is hidden with :attr:`hide_char`.
    hide_char : str, default: "*"
        Character to hide input when :attr:`hide_input` is true.
    max_chars : int | None, default: None
        Maximum allowed number of characters in textbox.
    alpha : float, default: 1.0
        Transparency of gadget.
    is_grabbable : bool, default: True
        Whether grabbable behavior is enabled.
    ptf_on_grab : bool, default: False
        Whether the gadget will be pulled to front when grabbed.
    mouse_button : MouseButton, default: "left"
        Mouse button used for grabbing.
    size : Size, default: Size(10, 10)
        Size of gadget.
    pos : Point, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: False
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    enter_callback : Callable[[Textbox], None] | None
        Called when textbox has focus and `enter` is pressed.
    placeholder : str
        Placeholder text for textbox.
    hide_input : bool
        Whether input is hidden with :attr:`hide_char`.
    hide_char : str
        Character to hide input when :attr:`hide_input` is true.
    max_chars : int | None
        Maximum allowed number of characters in textbox.
    alpha : float
        Transparency of gadget.
    text : str
        The textbox's text.
    cursor : int
        The cursor column.
    is_selecting : bool
        Whether there is a selection.
    has_nonempty_selection : bool
        Whether selection is non-empty.
    is_focused : bool
        Whether gadget has focus.
    any_focused : bool
        Whether any gadget has focus.
    is_grabbable : bool
        Whether grabbable behavior is enabled.
    ptf_on_grab : bool
        Whether the gadget will be pulled to front when grabbed.
    mouse_button : MouseButton
        Mouse button used for grabbing.
    is_grabbed : bool
        Whether gadget is grabbed.
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : SizeHint
        Size as a proportion of parent's height and width.
    pos_hint : PosHint
        Position as a proportion of parent's height and width.
    parent: Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App
        The running app.

    Methods
    -------
    undo()
        Undo previous edit.
    redo()
        Redo previous undo.
    select()
        Start a new selection at cursor if none.
    unselect()
        Unselect current selection.
    delete_selection()
        Delete current selection.
    move_cursor_left(n)
        Move cursor left `n` characters.
    move_cursor_right(n)
        Move cursor right `n` characters.
    move_word_left()
        Move cursor a word left.
    move_word_right()
        Move cursor a word right.
    update_theme()
        Paint the gadget with current theme.
    focus()
        Focus gadget.
    blur()
        Un-focus gadget.
    focus_next()
        Focus next focusable gadget.
    focus_previous()
        Focus previous focusable gadget.
    on_focus()
        Update gadget when it gains focus.
    on_blur()
        Update gadget when it loses focus.
    grab(mouse_event)
        Grab the gadget.
    ungrab(mouse_event)
        Ungrab the gadget.
    grab_update(mouse_event)
        Update gadget with incoming mouse events while grabbed.
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(\*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    def __init__(
        self,
        *,
        enter_callback: Callable[[Textbox], None] | None = None,
        placeholder: str = "",
        hide_input: bool = False,
        hide_char: NDArray[Cell] | str = "*",
        max_chars: int | None = None,
        is_grabbable: bool = True,
        ptf_on_grab: bool = False,
        mouse_button: MouseButton = "left",
        alpha: float = 1.0,
        size: Size = Size(10, 10),
        pos: Point = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = False,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        self._placeholder_gadget = Text(alpha=0.0, is_transparent=is_transparent)
        self._placeholder_gadget.set_text(placeholder)
        self._cursor = _Cursor()
        self._box = _Box(size=size, is_transparent=is_transparent)
        super().__init__(
            is_grabbable=is_grabbable,
            ptf_on_grab=ptf_on_grab,
            mouse_button=mouse_button,
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )

        self._selection_start = self._selection_end = None
        self._line_length = 0
        self._undo_stack = []
        self._redo_stack = []
        self._undo_buffer = []
        self._undo_buffer_type = "add"

        self._box.add_gadgets(self._placeholder_gadget, self._cursor)
        self.add_gadgets(self._box)

        self.enter_callback = enter_callback
        """Called when textbox has focus and `enter` is pressed."""
        self.placeholder = placeholder
        """Placeholder text for textbox."""
        self.hide_input = hide_input
        """Whether input is hidden with :attr:`hide_char`."""
        self.hide_char = hide_char
        """Character to hide input if :attr:`hide_input` is true."""
        self.max_chars = max_chars
        """Maximum allowed number of characters in textbox."""
        self.alpha = alpha

    @property
    def alpha(self) -> float:
        """Transparency of gadget."""
        return self._box.alpha

    @alpha.setter
    def alpha(self, alpha: float):
        self._box.alpha = alpha

    def on_transparency(self) -> None:
        """Update gadget after transparency is enabled/disabled."""
        self._box.is_transparent = self.is_transparent
        self._placeholder_gadget.is_transparent = self.is_transparent

    @property
    def hide_char(self) -> str:
        """Character to hide input when :attr:`hide_input` is true."""
        return self._hide_char

    @hide_char.setter
    def hide_char(self, char: str):
        self._hide_char = char[:1] or "*"

    def update_theme(self):
        """Paint the gadget with current theme."""
        primary = self.color_theme.textbox_primary
        fg = primary.fg
        bg = primary.bg
        self._box.canvas["fg_color"] = self._box.default_fg_color = fg
        self._box.canvas["bg_color"] = self._box.default_bg_color = bg
        self._cursor.fg_color = bg
        self._cursor.bg_color = fg

        placeholder = self.color_theme.textbox_placeholder
        self._placeholder_gadget.default_fg_color = placeholder.fg
        self._placeholder_gadget.default_bg_color = placeholder.bg
        self._placeholder_gadget.canvas["fg_color"] = placeholder.fg
        self._placeholder_gadget.canvas["bg_color"] = placeholder.bg

        self._highlight_selection()

    def on_size(self):
        """Resize and reposition children on resize."""
        self._box.width = max(self.width, self._line_length + 1)
        if self._box.x + self._line_length < self.width:
            self._box.x = min(0, self.width - self._line_length)
        self.cursor = self.cursor

    def on_focus(self):
        """Show cursor and select all text on focus."""
        self._cursor.is_enabled = True
        if self._line_length > 0:
            self._ctrl_a()

    def on_blur(self):
        """Hide cursor on blur."""
        self._cursor.is_enabled = False
        self.unselect()
        self.cursor = self.cursor

    @property
    def placeholder(self) -> str:
        """Placeholder text for textbox."""
        return self._placeholder

    @placeholder.setter
    def placeholder(self, placeholder: str):
        self._placeholder = placeholder
        self._placeholder_gadget.set_text(placeholder)
        self._placeholder_gadget.is_enabled = self._line_length == 0 and bool(
            placeholder
        )

    def _move_undo_buffer_to_stack(self, buffer_type=None):
        self._undo_buffer_type = buffer_type
        if self._undo_buffer:
            self._undo_stack.append(self._undo_buffer)
            self._undo_buffer = []
            self._redo_stack.clear()

    def undo(self):
        """Undo previous edit."""
        self._move_undo_buffer_to_stack()
        if self._undo_stack:
            redo = []
            for func, args, selection_start, selection_end, cursor in reversed(
                self._undo_stack.pop()
            ):
                redo.append(func(*args))
                self._selection_start = selection_start
                self._selection_end = selection_end
                self.cursor = cursor
            self._redo_stack.append(redo)

    def redo(self):
        """Redo previous undo."""
        if self._redo_stack and not self._undo_buffer:
            undo = []
            for func, args, selection_start, selection_end, cursor in reversed(
                self._redo_stack.pop()
            ):
                undo.append(func(*args))
                self._selection_start = selection_start
                self._selection_end = selection_end
                self.cursor = cursor
            self._undo_stack.append(undo)

    @property
    def text(self) -> str:
        """The textbox's text."""
        return "".join(self._box.canvas["char"][0, : self._line_length])

    @text.setter
    def text(self, text: str):
        text = text.replace("\n", " ")[: self.max_chars]
        self.unselect()
        self._del_text(0, self._line_length)
        self._add_text(0, text)
        self._redo_stack.clear()
        self._undo_stack.clear()
        self._undo_buffer.clear()
        self.cursor = self._line_length

    @property
    def cursor(self) -> int:
        """The cursor column."""
        return self._cursor.x

    @cursor.setter
    def cursor(self, cursor: int):
        """After setting cursor position, move textbox so that cursor is visible."""
        self._cursor.x = cursor
        self._placeholder_gadget.is_enabled = self._line_length == 0 and bool(
            self._placeholder
        )

        max_x = self.width - 1
        rel_x = cursor + self._box.x
        if rel_x > max_x:
            self._box.x -= rel_x - max_x
        elif rel_x < 0:
            self._box.x -= rel_x

        if self.is_selecting:
            self._selection_end = self.cursor

        self._highlight_selection()

    def _highlight_selection(self):
        colors = self._box.canvas[["fg_color", "bg_color"]]
        colors[:] = self._box.default_fg_color, self._box.default_bg_color

        if self._selection_start != self._selection_end:
            if self._selection_start > self._selection_end:
                start = self._selection_end
                end = self._selection_start
            else:
                start = self._selection_start
                end = self._selection_end

            colors[0, start:end] = self.color_theme.textbox_selection_highlight

    @property
    def is_selecting(self) -> bool:
        """Whether there is a selection."""
        return self._selection_start is not None and self._selection_end is not None

    @property
    def has_nonempty_selection(self) -> bool:
        """Whether selection is non-empty."""
        return self.is_selecting and self._selection_start != self._selection_end

    def select(self):
        """Start a new selection at cursor if none."""
        if not self.is_selecting:
            self._selection_start = self._selection_end = self.cursor

    def unselect(self):
        """Unselect current selection."""
        self._selection_start = self._selection_end = None

    def delete_selection(self):
        """Delete current selection."""
        if self.has_nonempty_selection:
            return self._del_text(self._selection_start, self._selection_end)

    def _del_text(self, start: int, end: int):
        if start > end:
            start, end = end, start

        if end > self._line_length:
            # ! If we ended up here, something went wrong.
            end = self._line_length

        contents = "".join(self._box.canvas["char"][0, start:end])
        selection_start = self._selection_start
        selection_end = self._selection_end
        cursor = self.cursor

        len_end = self._line_length - end
        self._line_length = start + len_end

        box = self._box
        box.canvas[0, start : self._line_length] = box.canvas[0, end : end + len_end]
        box.canvas[0, self._line_length :] = box.default_cell

        self.unselect()
        self.cursor = start

        return self._add_text, [start, contents], selection_start, selection_end, cursor

    def _add_text(self, x: int, text: str):
        selection_start = self._selection_start
        selection_end = self._selection_end
        cursor = self.cursor

        text = text.replace("\n", " ")
        box = self._box
        box_text = (
            f"{''.join(box.canvas['char'][0, :x])}"
            f"{text}"
            f"{''.join(box.canvas['char'][0, x : self._line_length])}"
        )[: self.max_chars]

        box_width = self._line_length = str_width(box_text)
        if box_width >= box.width:
            box.width = box_width + 1

        box.add_str(box_text)
        box.canvas[0, box_width:] = box.default_cell

        self.cursor = min(box_width, x + str_width(text))
        return self._del_text, [x, self.cursor], selection_start, selection_end, cursor

    def move_cursor_left(self, n: int = 1):
        """Move cursor left `n` characters."""
        text_before_cursor = "".join(self._box.canvas["char"][0, : self.cursor])
        nchars_before_cursor = len(text_before_cursor)
        if n <= nchars_before_cursor:
            self.cursor = str_width(text_before_cursor[:-n])
        else:
            self.cursor = 0

    def move_cursor_right(self, n: int = 1):
        """Move cursor right `n` characters."""
        text_after_cursor = "".join(
            self._box.canvas["char"][0, self.cursor : self._line_length]
        )
        nchars_after_cursor = len(text_after_cursor)
        if n <= nchars_after_cursor:
            self.cursor += str_width(text_after_cursor[:n])
        else:
            self.cursor = self._line_length

    def move_word_left(self):
        """Move cursor a word left."""
        last_x = self.cursor
        first_char_found = False
        while True:
            self.move_cursor_left()
            if self.cursor == last_x:
                break

            last_x = self.cursor

            current_char = self._box.canvas[0, self.cursor]["char"]
            if not first_char_found:
                if not current_char.isspace():
                    first_char_found = True
                    char_is_word_char = is_word_char(current_char)
            elif current_char.isspace() or char_is_word_char != is_word_char(
                current_char
            ):
                self.move_cursor_right()
                break

    def move_word_right(self):
        """Move cursor a word right."""
        last_x = self.cursor
        first_char_found = False
        while True:
            self.move_cursor_right()
            if self.cursor == last_x:
                break

            last_x = self.cursor

            current_char = self._box.canvas[0, self.cursor]["char"]
            if not first_char_found:
                if not current_char.isspace():
                    first_char_found = True
                    char_is_word_char = is_word_char(current_char)
            elif current_char.isspace() or char_is_word_char != is_word_char(
                current_char
            ):
                break

    def _enter(self):
        if self.enter_callback is not None:
            self.enter_callback(self)

    def _backspace(self):
        if self.has_nonempty_selection:
            self._move_undo_buffer_to_stack("del")
            self._undo_buffer.append(self.delete_selection())
        else:
            if self._undo_buffer_type != "del":
                self._move_undo_buffer_to_stack("del")

            end = self.cursor
            self.move_cursor_left()
            start = self.cursor
            self.cursor = end
            if start != end:
                self._undo_buffer.append(self._del_text(start, end))

    def _delete(self):
        if self.has_nonempty_selection:
            self._move_undo_buffer_to_stack("del")
            self._undo_buffer.append(self.delete_selection())
        else:
            if self._undo_buffer_type != "del":
                self._move_undo_buffer_to_stack("del")

            start = self.cursor
            self.move_cursor_right()
            end = self.cursor
            self.cursor = start
            if start != end:
                self._undo_buffer.append(self._del_text(start, end))

    def _left(self):
        if self.has_nonempty_selection:
            select_start = min(self._selection_start, self._selection_end)
            self.unselect()
            self.cursor = select_start
        else:
            self.unselect()
            self.move_cursor_left()

    def _right(self):
        if self.has_nonempty_selection:
            select_end = max(self._selection_start, self._selection_end)
            self.unselect()
            self.cursor = select_end
        else:
            self.unselect()
            self.move_cursor_right()

    def _ctrl_left(self):
        self.unselect()
        self.move_word_left()

    def _ctrl_right(self):
        self.unselect()
        self.move_word_right()

    def _ctrl_a(self):
        """Select all."""
        self._selection_start = 0
        self._selection_end = self._line_length
        self.cursor = self._line_length

    def _ctrl_d(self):
        """Select word."""
        self.unselect()
        last_x = self.cursor
        while True:
            self.move_cursor_left()
            if last_x == self.cursor:
                break
            if not is_word_char(self._box.canvas[0, self.cursor]["char"]):
                self.move_cursor_right()
                break
            last_x = self.cursor

        self.select()
        last_x = self.cursor
        while True:
            if not is_word_char(self._box.canvas[0, self.cursor]["char"]):
                break
            self.move_cursor_right()
            if last_x == self.cursor:
                break
            last_x = self.cursor

    def _home(self):
        self.unselect()
        self.cursor = 0

    def _end(self):
        self.unselect()
        self.cursor = self._line_length

    def _shift_left(self):
        self.select()
        self.move_cursor_left()

    def _shift_right(self):
        self.select()
        self.move_cursor_right()

    def _shift_ctrl_left(self):
        self.select()
        self.move_word_left()

    def _shift_ctrl_right(self):
        self.select()
        self.move_word_right()

    def _shift_home(self):
        self.select()
        self.cursor = 0

    def _shift_end(self):
        self.select()
        self.cursor = self._line_length

    def _escape(self):
        if self.has_nonempty_selection:
            self.unselect()
            self._highlight_selection()
        else:
            self.unselect()
            self.blur()

    def _ascii(self, key):
        if self.has_nonempty_selection:
            self._move_undo_buffer_to_stack("add")
            self._undo_buffer.append(self.delete_selection())

        if (
            self.max_chars is None
            or (self._box.canvas["char"][0, : self._line_length] != "").sum()
            < self.max_chars
        ):
            if self._undo_buffer_type != "add":
                self._move_undo_buffer_to_stack("add")
            self._undo_buffer.append(self._add_text(self.cursor, key))

    __HANDLERS = {
        ("enter", False, False, False): _enter,
        ("backspace", False, False, False): _backspace,
        ("delete", False, False, False): _delete,
        ("left", False, False, False): _left,
        ("right", False, False, False): _right,
        ("left", False, True, False): _ctrl_left,
        ("right", False, True, False): _ctrl_right,
        ("home", False, False, False): _home,
        ("end", False, False, False): _end,
        ("left", False, False, True): _shift_left,
        ("right", False, False, True): _shift_right,
        ("left", False, True, True): _shift_ctrl_left,
        ("right", False, True, True): _shift_ctrl_right,
        ("home", False, False, True): _shift_home,
        ("end", False, False, True): _shift_end,
        ("escape", False, False, False): _escape,
        ("z", False, True, False): undo,
        ("z", False, True, True): redo,
        ("y", False, True, False): redo,  # ctrl + shift + z won't work on linux
        ("r", False, True, False): redo,
        ("a", False, True, False): _ctrl_a,
        ("d", False, True, False): _ctrl_d,
    }

    def on_key(self, key_event: KeyEvent) -> bool | None:
        """Process key press."""
        if not self.is_focused:
            return super().on_key(key_event)

        if (
            not key_event.alt
            and not key_event.ctrl
            and not key_event.shift
            and len(key_event.key) == 1
        ):
            self._ascii(key_event.key)
        elif handler := self.__HANDLERS.get(astuple(key_event)):
            handler(self)
        else:
            return super().on_key(key_event)

        return True

    def on_paste(self, paste_event: PasteEvent) -> bool | None:
        """Add paste to textbox."""
        if not self.is_focused:
            return

        self._move_undo_buffer_to_stack()
        undos = []
        if undo := self.delete_selection():
            undos.append(undo)
        undos.append(self._add_text(self.cursor, paste_event.paste))
        self._undo_stack.append(undos)
        self._redo_stack.clear()

        return True

    def grab(self, mouse_event):
        """Start selection on grab."""
        if mouse_event.button == "left" and self._box.collides_point(mouse_event.pos):
            super().grab(mouse_event)

            _, x = self._box.to_local(mouse_event.pos)

            if not mouse_event.shift:
                self.unselect()

            self.cursor = min(x, self._line_length)
            self.select()  # Need at least an empty selection for `grab_update`.

    def grab_update(self, mouse_event: MouseEvent):
        """Update selection on grab update."""
        if self._box.collides_point(mouse_event.pos):
            _, x = self._box.to_local(mouse_event.pos)
            self.cursor = min(x, self._line_length)
        else:
            _, x = self.to_local(mouse_event.pos)

            if x < 0:
                self.move_cursor_left()
            elif x >= self.width:
                self.move_cursor_right()

    def ungrab(self, mouse_event):
        """Clear an empty selection on ungrab."""
        super().ungrab(mouse_event)
        if self._selection_start == self._selection_end:
            self.unselect()



r"""
Terminal rendering.

``render_root`` is responsible for generating ansi from the diffs of the root's double
buffer. Ansi is generated in a bit of a naive way, where the cursor is moved to specific
coordinate then all sgr parameters of the cell is output, for each cell in the diff.
This generates a lot of ansi for large diffs. In the future, this function might decide
to try and output as little ansi as possible, if there is a noticable performance gain
to do so.
"""

import numpy as np

from .gadgets._root import _Root
from .terminal import Vt100Terminal
from .text_tools import char_width


def render_root(root: _Root, terminal: Vt100Terminal) -> None:
    """
    Render root canvas into a terminal.

    Parameters
    ----------
    root : _Root
        Root gadget of gadget tree.
    terminal : Vt100Terminal
        A VT100 terminal.
    """
    if terminal.expect_dsr():
        return

    w = root.width
    inline_column = root.pos.x
    write = terminal._out_buffer.append
    inline = not terminal.in_alternate_screen
    last_y = 0

    resized = root._resized  # ! Must grab before calling root._render
    root._render()
    canvas = root.canvas  # ! Must grab after calling root._render

    if resized:
        ys, xs = np.indices(root.size).reshape(2, -1)
    else:
        diffs = root._last_canvas != canvas
        ys, xs = diffs.nonzero()

    # Save cursor
    write("\x1b7")
    if inline:
        terminal.move_cursor(root._pos)
    for y, x, cell in zip(ys, xs, canvas[ys, xs]):
        (
            char,
            bold,
            italic,
            underline,
            strikethrough,
            overline,
            reverse,
            (fr, fg, fb),  # foreground color
            (br, bg, bb),  # background color
        ) = cell.item()

        # The following conditions ensure full-width glyphs "have enough room" else
        # they are not painted.
        if char == "":
            # `""` is used to indicate the character before it is a full-width
            # character. If this char is appearing in the diffs, we probably need to
            # repaint the full-width character before it, but if the character
            # before it isn't full-width paint whitespace instead.
            if x > 0 and char_width(canvas["char"][y, x - 1].item()) == 2:
                x -= 1
                (
                    char,
                    bold,
                    italic,
                    underline,
                    strikethrough,
                    overline,
                    reverse,
                    (fr, fg, fb),
                    (br, bg, bb),
                ) = canvas[y, x].item()
            else:
                char = " "
        elif (
            x + 1 < w
            and canvas["char"][y, x + 1].item() != ""
            and char_width(char) == 2
        ):
            # If the character is full-width, but the following character isn't
            # `""`, assume the full-width character is being clipped, and paint
            # whitespace instead.
            char = " "

        if inline:
            # Note that `y`s are non-decreasing.
            if last_y < y:
                # Move down `y - last_y` rows.
                write(f"\x1b[{y - last_y}B")
            # Move to column `x + 1`.
            write(f"\x1b[{inline_column + x + 1}G")
        else:
            # Move cursor to position `(y + 1, x + 1)`.
            write(f"\x1b[{y + 1};{x + 1}H")
        last_y = y

        write(
            "\x1b[0;"  # Reset attributes.
            f"{'1;' if bold else ''}"
            f"{'3;' if italic else ''}"
            f"{'4;' if underline else ''}"
            f"{'9;' if strikethrough else ''}"
            f"{'53;' if overline else ''}"
            f"{'7;' if reverse else ''}"
            f"38;2;{fr};{fg};{fb};48;2;{br};{bg};{bb}m"  # Set color pair.
            f"{char}"
        )
    # Restore cursor
    write("\x1b8")
    terminal.flush()



"""Commonly used colors."""

from pygments.style import Style
from pygments.token import (
    Comment,
    Error,
    Generic,
    Keyword,
    Literal,
    Name,
    Number,
    Operator,
    String,
    Text,
    Token,
)

from .color_types import AColor, Color, ColorTheme

__all__ = [
    "ABLACK",
    "ABLUE",
    "ACYAN",
    "AGREEN",
    "AMAGENTA",
    "ARED",
    "AWHITE",
    "AYELLOW",
    "BLACK",
    "BLUE",
    "CYAN",
    "DEFAULT_COLOR_THEME",
    "DEFAULT_PRIMARY_BG",
    "DEFAULT_PRIMARY_FG",
    "GREEN",
    "MAGENTA",
    "RED",
    "TRANSPARENT",
    "WHITE",
    "YELLOW",
    "Neptune",
]

WHITE = Color(255, 255, 255)
"""White."""

BLACK = Color(0, 0, 0)
"""Black."""

RED = Color(255, 0, 0)
"""Red."""

GREEN = Color(0, 255, 0)
"""Green."""

BLUE = Color(0, 0, 255)
"""Blue."""

YELLOW = Color(255, 255, 0)
"""Yellow."""

CYAN = Color(0, 255, 255)
"""Cyan."""

MAGENTA = Color(255, 0, 255)
"""Magenta."""

AWHITE = AColor(255, 255, 255)
"""Opaque white."""

ABLACK = AColor(0, 0, 0)
"""Opaque black."""

ARED = AColor(255, 0, 0)
"""Opaque red."""

AGREEN = AColor(0, 255, 0)
"""Opaque green."""

ABLUE = AColor(0, 0, 255)
"""Opaque blue."""

AYELLOW = AColor(255, 255, 0)
"""Opaque yellow."""

ACYAN = AColor(0, 255, 255)
"""Opaque cyan."""

AMAGENTA = AColor(255, 0, 255)
"""Opaque magenta."""

TRANSPARENT = AColor(0, 0, 0, 0)
"""Transparent black."""

DEFAULT_COLOR_THEME: ColorTheme = {
    "primary": {"fg": "f6a7a9", "bg": "070c25"},
    "text_pad_line_highlight": {"fg": "f6a7a9", "bg": "0c0e30"},
    "text_pad_selection_highlight": {"fg": "f6a7a9", "bg": "0f1847"},
    "textbox_primary": {"fg": "fff0f6", "bg": "070c25"},
    "textbox_selection_highlight": {"fg": "fff0f6", "bg": "0f1847"},
    "textbox_placeholder": {"fg": "2a3a92", "bg": "070c25"},
    "button_normal": {"fg": "dde4ed", "bg": "2a3ca0"},
    "button_hover": {"fg": "fff0f6", "bg": "3248c0"},
    "button_press": {"fg": "fff0f6", "bg": "c4a219"},
    "button_disallowed": {"fg": "272b40", "bg": "070c25"},
    "menu_item_hover": {"fg": "f2babc", "bg": "111834"},
    "menu_item_selected": {"fg": "ecf3ff", "bg": "1b244b"},
    "menu_item_disallowed": {"fg": "272b40", "bg": "070c25"},
    "titlebar_normal": {"fg": "ffe0df", "bg": "070c25"},
    "titlebar_inactive": {"fg": "7d6b71", "bg": "070c25"},
    "data_table_sort_indicator": {"fg": "ecf3ff", "bg": "070c25"},
    "data_table_hover": {"fg": "f6a7a9", "bg": "111834"},
    "data_table_stripe": {"fg": "f6a7a9", "bg": "0b1238"},
    "data_table_stripe_hover": {"fg": "f6a7a9", "bg": "0f184a"},
    "data_table_selected": {"fg": "ecf3ff", "bg": "111f5e"},
    "data_table_selected_hover": {"fg": "ecf3ff", "bg": "1b244b"},
    "progress_bar": {"fg": "ffe0df", "bg": "2a3ca0"},
    "markdown_link": {"fg": "376cff", "bg": "070c25"},
    "markdown_link_hover": {"fg": "4668ff", "bg": "070c25"},
    "markdown_inline_code": {"fg": "806ae5", "bg": "080b1a"},
    "markdown_quote": {"fg": "2054e2", "bg": "0c1b4b"},
    "markdown_title": {"fg": "cfd1d4", "bg": "292a2d"},
    "markdown_image": {"fg": "f6a7a9", "bg": "0c1540"},
    "markdown_block_code_background": "080b1a",
    "markdown_quote_block_code_background": "11265d",
    "markdown_header_background": "030612",
    "scroll_view_scrollbar": "070c25",
    "scroll_view_indicator_normal": "0e1843",
    "scroll_view_indicator_hover": "111e4f",
    "scroll_view_indicator_press": "172868",
    "window_border_normal": "122162",
    "window_border_inactive": "282c3e",
}
"""Default color theme."""

DEFAULT_PRIMARY_FG = Color.from_hex("f6a7a9")
"""Default primary foreground color."""

DEFAULT_PRIMARY_BG = Color.from_hex("070c25")
"""Default primary background color."""


class Neptune(Style):
    """
    Default syntax highlight color theme.

    This style is adapted from
    https://github.com/yl92/paddy-color-theme/blob/master/themes/upright/Paddy-neptune-upright-color-theme.json.
    """

    name = "neptune"

    background_color = "#070c25"
    highlight_color = "#5f32db"

    line_number_special_color = "#2f3adf"
    line_number_special_background_color = "#5f32db"

    line_number_color = "#527bff"
    line_number_background_color = "#070c25"

    styles = {
        Token: "#f6a7a9",
        Error: "bold #d32e11",
        Keyword: "#5bb177",
        Keyword.Constant: "bold #91a5bd",
        Keyword.Pseudo: "#e28f37",
        Name: "#f6a7a9",
        Name.Class: "bold #de9138",
        Name.Constant: "#17c78c",
        Name.Decorator: "#d85757",
        Name.Exception: "bold #de9138",
        Name.Function: "#d85757",
        Name.Label: "#f6a7a9",
        Name.Property: "#f6a7a9",
        Name.Tag: "bold #309388",
        Name.Variable: "#f6a7a9",
        Literal: "#da7763",
        String: "#da7763",
        String.Affix: "#c75974",
        String.Escape: "#62A594",
        String.Regex: "#d46fbe",
        Number: "#e47559",
        Comment: "#554492",
        Comment.Special: "#978d7a",
        Operator: "#5bb177",
        Generic: "#f6a7a9",
        Generic.Deleted: "bold bg:#cf5076 #309388",
        Generic.Emph: "italic",
        Generic.Error: "bold #d32e11",
        Generic.Heading: "#da5656",
        Generic.Inserted: "bg:#4d67ff #da7763",
        Generic.Output: "#69a097",
        Generic.Strong: "bold",
        Generic.EmphStrong: "bold italic",
        Generic.Subheading: "#da5656",
        Generic.Traceback: "bold #de9138",
        Generic.Underline: "underline",
        Text: "#f6a7a9",
    }



"""Themable behavior for gadgets."""

from abc import ABC, abstractmethod
from dataclasses import dataclass, fields
from typing import NamedTuple

from ...colors import DEFAULT_COLOR_THEME, Color, ColorTheme

__all__ = ["Themable"]


class _ColorPair(NamedTuple):
    """A foreground and background pair of colors."""

    fg: Color
    """The foreground color."""
    bg: Color
    """The background color."""


@dataclass
class _ColorTheme:
    def __init__(self, color_theme: ColorTheme):
        for field in fields(self):
            name = field.name
            info = color_theme.get(name, DEFAULT_COLOR_THEME[name])

            if isinstance(info, str):
                color = Color.from_hex(info)
            else:
                color = _ColorPair(
                    Color.from_hex(info["fg"]), Color.from_hex(info["bg"])
                )
            setattr(self, name, color)

    primary: _ColorPair
    """The primary color pair."""
    text_pad_line_highlight: _ColorPair
    """Text pad line highlight color pair."""
    text_pad_selection_highlight: _ColorPair
    """Text pad selection color pair."""
    textbox_primary: _ColorPair
    """Text pad primary color pair."""
    textbox_selection_highlight: _ColorPair
    """Textbox selection color pair."""
    textbox_placeholder: _ColorPair
    """Textbox placeholder color pair."""
    button_normal: _ColorPair
    """Button normal color pair."""
    button_hover: _ColorPair
    """Button hover color pair."""
    button_press: _ColorPair
    """Button press color pair."""
    button_disallowed: _ColorPair
    """Button disallowed color pair."""
    menu_item_hover: _ColorPair
    """Menu item hover color pair."""
    menu_item_selected: _ColorPair
    """Menu item selected color pair."""
    menu_item_disallowed: _ColorPair
    """Menu item disallowed color pair."""
    titlebar_normal: _ColorPair
    """Titlebar normal color pair."""
    titlebar_inactive: _ColorPair
    """Titlebar inactive color pair."""
    data_table_sort_indicator: _ColorPair
    """Data table sort indicator color pair."""
    data_table_hover: _ColorPair
    """Data table hover color pair."""
    data_table_stripe: _ColorPair
    """Data table stripe color pair."""
    data_table_stripe_hover: _ColorPair
    """Data table stripe hover color pair."""
    data_table_selected: _ColorPair
    """Data table selected color pair."""
    data_table_selected_hover: _ColorPair
    """Data table selected hover color pair."""
    progress_bar: _ColorPair
    """Progress bar color pair."""
    markdown_link: _ColorPair
    """Markdown link color pair."""
    markdown_link_hover: _ColorPair
    """Markdown link hover color pair."""
    markdown_inline_code: _ColorPair
    """Markdown inline code color pair."""
    markdown_quote: _ColorPair
    """Markdown quote color pair."""
    markdown_title: _ColorPair
    """Markdown title color pair."""
    markdown_image: _ColorPair
    """Markdown image color pair."""
    markdown_block_code_background: Color
    """Markdown block code background color."""
    markdown_quote_block_code_background: Color
    """Markdown quote block code background color."""
    markdown_header_background: Color
    """Markdown header background color."""
    scroll_view_scrollbar: Color
    """Scroll view scrollbar color."""
    scroll_view_indicator_normal: Color
    """Scroll view indicator normal color."""
    scroll_view_indicator_hover: Color
    """Scroll view indicator hover color."""
    scroll_view_indicator_press: Color
    """Scroll view indicator press color."""
    window_border_normal: Color
    """Window border normal color."""
    window_border_inactive: Color
    """Window border inactive color."""


class Themable(ABC):
    """
    Themable behavior for a gadget.

    Themable gadgets share a color theme. They must implement :meth:`update_theme`
    which paints the gadget with current theme.

    Whenever the running app's theme is changed, `update_theme` will be called
    for all :class:`Themable` gadgets.

    Methods
    -------
    update_theme()
        Paint the gadget with current theme.
    """

    color_theme: _ColorTheme

    @classmethod
    def set_theme(cls, color_theme: ColorTheme):
        """Set color theme."""
        cls.color_theme = _ColorTheme(color_theme)

    def on_add(self):
        """Update theme."""
        super().on_add()
        self.update_theme()

    @abstractmethod
    def update_theme(self):
        """Paint the gadget with current theme."""



"""Build cython extensions and generated files."""

import sys

from Cython.Build import cythonize
from setuptools import setup
from setuptools.command.build_py import build_py
from wcwidth import wcwidth

_CHAR_WIDTHS_DOC = '''\
"""
Character widths as generated by wcwidth.

Each tuple in `CHAR_WIDTHS` represents an interval of ords with common width, i.e.,
`(0, 31, 0)` means ords 0 through 31 (inclusive) have width 0. Intervals with width 1
are omitted so if an ord doesn't belong to any interval we can assume it has width 1.
"""

'''


def _create_char_widths():
    """Build ``_char_widths.py``."""
    groups = []
    start = 0
    group_width = 1
    for codepoint in range(sys.maxunicode + 1):
        char_width = max(wcwidth(chr(codepoint)), 0)
        if char_width != group_width:
            if group_width != 1:
                groups.append((start, codepoint - 1, group_width))
            group_width = char_width
            start = codepoint

    if group_width != 1:
        groups.append((start, codepoint - 1, group_width))

    with open("src/batgrl/_char_widths.py", "w") as file:
        file.write(_CHAR_WIDTHS_DOC)
        file.write("CHAR_WIDTHS = (\n")
        for group in groups:
            file.write(f"    {group},\n")
        file.write(")\n")


class build_py_with_char_widths(build_py):
    """Generate ``_char_widths.py`` on build."""

    def run(self):
        """Generate ``_char_widths.py`` on build."""
        super().run()
        _create_char_widths()


setup(
    ext_modules=cythonize(["src/batgrl/geometry/regions.pyx"]),
    cmdclass={"build_py": build_py_with_char_widths},
)



"""A linux VT100 terminal."""

import asyncio
import os
import select
import signal
import sys
import termios
import tty
from collections.abc import Callable
from typing import Final

from .events import Event, ResizeEvent
from .vt100_terminal import Vt100Terminal

STDIN: Final = sys.stdin.fileno()


class LinuxTerminal(Vt100Terminal):
    """
    A linux VT100 terminal.

    Attributes
    ----------
    in_alternate_screen : bool
        Whether the alternate screen buffer is enabled.

    Methods
    -------
    process_stdin()
        Read from stdin and feed data into input parser to generate events.
    raw_mode()
        Set terminal to raw mode.
    restore_console()
        Restore console to its original mode.
    attach(event_handler)
        Start generating events from stdin.
    unattach()
        Stop generating events from stdin.
    events()
        Return a list of input events and reset the event buffer.
    get_size()
        Get terminal size.
    flush()
        Write buffer to output stream and flush.
    set_title(title)
        Set terminal title.
    enter_alternate_screen()
        Enter alternate screen buffer.
    exit_alternate_screen()
        Exit alternate screen buffer.
    enable_mouse_support()
        Enable mouse support in terminal.
    disable_mouse_support()
        Disable mouse support in terminal.
    reset_attributes()
        Reset character attributes.
    enable_bracketed_paste()
        Enable bracketed paste in terminal.
    disable_bracketed_paste()
        Disable bracketed paste in terminal.
    show_cursor()
        Show cursor in terminal.
    hide_cursor()
        Hide cursor in terminal.
    enable_reporting_focus()
        Enable reporting terminal focus.
    disable_reporting_focus()
        Disable reporting terminal focus.
    request_cursor_position_report()
        Report current cursor position.
    request_foreground_color()
        Report terminal foreground color.
    request_background_color()
        Report terminal background color.
    request_device_attributes()
        Report device attributes.
    expect_dsr()
        Return whether a device status report is expected.
    move_cursor(pos)
        Move cursor to ``pos``.
    erase_in_display(n)
        Clear part of the screen.
    """

    def process_stdin(self) -> None:
        """Read from stdin and feed data into input parser to generate events."""
        reads = []
        while select.select([STDIN], [], [], 0)[0]:
            try:
                read = os.read(STDIN, 1024)
            except OSError:
                break
            else:
                reads.append(read)

        data = b"".join(reads).decode(errors="surrogateescape")
        self._feed(data)

    def raw_mode(self) -> None:
        """Set terminal to raw mode."""
        self._original_mode = termios.tcgetattr(STDIN)
        attrs_raw = termios.tcgetattr(STDIN)
        attrs_raw[tty.LFLAG] &= ~(
            termios.ECHO | termios.ICANON | termios.IEXTEN | termios.ISIG
        )
        attrs_raw[tty.IFLAG] &= ~(
            termios.IXON | termios.IXOFF | termios.ICRNL | termios.INLCR | termios.IGNCR
        )
        attrs_raw[tty.CC][termios.VMIN] = 1
        termios.tcsetattr(STDIN, termios.TCSANOW, attrs_raw)

    def restore_console(self) -> None:
        """Restore console to its original mode."""
        termios.tcsetattr(STDIN, termios.TCSANOW, self._original_mode)
        del self._original_mode

    def attach(self, event_handler: Callable[[list[Event]], None]) -> None:
        """
        Start generating events from stdin.

        Parameters
        ----------
        event_handler : Callable[[list[Event]], None]
            Callable that handles input events.
        """
        self._event_buffer.clear()
        self._event_handler = event_handler

        def process():
            self.process_stdin()
            event_handler(self.events())

        loop = asyncio.get_running_loop()
        loop.add_reader(STDIN, process)

        def on_resize(*_):
            self._event_buffer.append(ResizeEvent(self.get_size()))
            loop.call_soon_threadsafe(process)

        signal.signal(signal.SIGWINCH, on_resize)

    def unattach(self) -> None:
        """Stop generating events from stdin."""
        self._event_handler = None
        loop = asyncio.get_running_loop()
        loop.remove_reader(STDIN)
        signal.signal(signal.SIGWINCH, signal.SIG_DFL)



"""Base class for all gadgets."""

from __future__ import annotations

import asyncio
from collections.abc import Callable, Iterator, MutableSequence, Sequence
from functools import wraps
from itertools import count
from numbers import Real
from time import perf_counter
from types import MappingProxyType
from typing import Final, Literal, Self, TypedDict
from weakref import WeakKeyDictionary

import numpy as np
from numpy.typing import NDArray

from ..geometry import EASINGS, Easing, Point, Region, Size, clamp, lerp, round_down
from ..terminal.events import FocusEvent, KeyEvent, MouseEvent, PasteEvent
from ..text_tools import Cell, new_cell

__all__ = [
    "Anchor",
    "Cell",
    "Easing",
    "Point",
    "PosHint",
    "Region",
    "Size",
    "SizeHint",
    "Gadget",
    "bindable",
    "new_cell",
    "clamp",
    "lerp",
]

_UID: Final = count(1)
_ANCHOR_TO_POS: Final = {
    "top-left": (0.0, 0.0),
    "top": (0.0, 0.5),
    "top-right": (0.0, 1.0),
    "left": (0.5, 0.0),
    "center": (0.5, 0.5),
    "right": (0.5, 1.0),
    "bottom-left": (1.0, 0.0),
    "bottom": (1.0, 0.5),
    "bottom-right": (1.0, 1.0),
}
_DEFAULT_POS_HINT: Final = {
    "anchor": "center",
    "y_hint": None,
    "x_hint": None,
    "x_offset": 0,
    "y_offset": 0,
}
"""The default pos hint."""
_DEFAULT_SIZE_HINT: Final = {
    "height_hint": None,
    "width_hint": None,
    "height_offset": 0,
    "width_offset": 0,
    "max_height": None,
    "min_height": None,
    "max_width": None,
    "min_width": None,
}
"""The default size hint."""

Anchor = Literal[
    "top-left",
    "top",
    "top-right",
    "left",
    "center",
    "right",
    "bottom-left",
    "bottom",
    "bottom-right",
]
"""Point of gadget attached to a pos hint."""


class _GadgetList(MutableSequence):
    """
    A sequence of sibling gadgets.

    Gadget regions are invalidated if gadgets later in the sequence are added or
    removed.
    """

    def __init__(self) -> None:
        self._gadgets: list[Gadget] = []

    def __len__(self) -> int:
        return len(self._gadgets)

    def __getitem__(self, index: int) -> Gadget:
        return self._gadgets[index]

    def __setitem__(self, *_) -> None:
        raise NotImplementedError("_GadgetList.__setitem__ not implemented.")

    def __delitem__(self, index: int) -> None:
        del self._gadgets[index]
        for i in range(index - 1, -1, -1):
            self[i]._invalidate_region()

    def insert(self, index: int, gadget: Gadget) -> None:
        self._gadgets.insert(index, gadget)
        for i in range(index, -1, -1):
            self[i]._invalidate_region()

    def __iter__(self) -> Iterator[Gadget]:
        return iter(self._gadgets)

    def copy(self) -> list[Gadget]:
        """
        Return a copy of the internal list of gadgets.

        Note this doesn't return a ``_GadgetList``.
        """
        return self._gadgets.copy()


class PosHint(TypedDict, total=False):
    """
    A position hint.

    A pos hint allows a gadget to automatically position itself when added to the
    gadget tree or when its parent resizes. ``y_hint`` controls vertical position and
    ``x_hint`` horizontal. ``anchor`` determines which point of the gadget is attached
    to the pos hint. For instance, in the diagram below, if the ``y_hint`` and
    ``x_hint`` are both ``0.5`` the pos hint will be at point ``c`` on the parent
    (50% of the parent's height and width). Additionally, if the anchor is ``"top"``,
    the anchor will be at point ``a`` on the gadget::

             parent
        +---------------+
        |               |    +---a---+
        |       c       |    |gadget |
        |               |    +-------+
        +---------------+


    Subsequently, ``a`` will be aligned with ``c``, so that the gadget is positioned as
    below::

             parent
        +---------------+
        |               |
        |   +-------+   |
        |   |gadget |   |
        +---+-------+---+

    The additional parameters ``y_offset`` and ``x_offset`` allow one to translate the
    gadget by some integer offsets after the pos hint has been applied.

    Attributes
    ----------
    anchor : Anchor | tuple[float, float]
        Determines which point is attached to the pos hint.
    y_hint : float | None
        Vertical position as a proportion of parent's height.
    x_hint : float | None
        Horizontal position as a proportion of parent's width.
    y_offset : int
        Vertical offset after pos hint is applied.
    x_offset : int
        Horizontal offset after pos hint is applied.
    """

    anchor: Anchor | tuple[float, float]
    """Determines which point is attached to the pos hint."""
    y_hint: float | None
    """Vertical position as a proportion of parent's height."""
    x_hint: float | None
    """Horizontal position as a proportion of parent's width."""
    y_offset: int
    """Vertical offset after pos hint is applied."""
    x_offset: int
    """Horizontal offset after pos hint is applied."""


class SizeHint(TypedDict, total=False):
    """
    A size hint.

    A size hint allows a gadget to automatically size itself when added to the gadget
    tree or when its parent resizes. ``height_hint`` is the proportion of the parent's
    height the gadget's height will be and ``width_hint`` the proportion of the parent's
    width. Additional parameters ``height_offset`` and ``width_offset`` allow adjusting
    the size by some integer amount after the size hint has been applied. If given,
    ``min_height``, ``max_height``, ``min_width``, and ``max_width`` will prevent the
    gadget from sizing too small or too large.

    Attributes
    ----------
    height_hint : float | None
        Height as a proportion of parent's height.
    width_hint : float | None
        Width as a proportion of parent's width.
    height_offset : int
        Height offset after height-hint is applied.
    width_offset : int
        Width offset after width-hint is applied.
    max_height : int | None
        Maximum allowed height.
    min_height : int | None
        Minimum allowed height.
    max_width : int | None
        Maximum allowed width.
    min_width : int | None
        Minimum allowed width.
    """

    height_hint: float | None
    """Height as a proportion of parent's height."""
    width_hint: float | None
    """Width as a proportion of parent's width."""
    height_offset: int
    """Height offset after height-hint is applied."""
    width_offset: int
    """Width offset after width-hint is applied."""
    max_height: int | None
    """Maximum allowed height."""
    min_height: int | None
    """Minimum allowed height."""
    max_width: int | None
    """Maximum allowed width."""
    min_width: int | None
    """Minimum allowed width."""


def _normalize_pos_hint(pos_hint: PosHint) -> PosHint:
    normal_hint = _DEFAULT_POS_HINT | pos_hint
    if normal_hint["y_hint"] is not None:
        normal_hint["y_hint"] = float(normal_hint["y_hint"])
    if normal_hint["x_hint"] is not None:
        normal_hint["x_hint"] = float(normal_hint["x_hint"])
    if isinstance(normal_hint["anchor"], str):
        normal_hint["anchor"] = _ANCHOR_TO_POS[normal_hint["anchor"]]
    return normal_hint


def _normalize_size_hint(size_hint: SizeHint) -> SizeHint:
    normal_hint = _DEFAULT_SIZE_HINT | size_hint
    if normal_hint["height_hint"] is not None:
        normal_hint["height_hint"] = float(normal_hint["height_hint"])
    if normal_hint["width_hint"] is not None:
        normal_hint["width_hint"] = float(normal_hint["width_hint"])
    return normal_hint


def bindable(setter):
    """Decorate property setters to make them bindable."""
    instances: WeakKeyDictionary[Gadget, Callable[[], None]] = WeakKeyDictionary()

    @wraps(setter)
    def wrapper(self, *args, **kwargs):
        setter(self, *args, **kwargs)
        if bindings := instances.get(self):
            for callback in bindings.values():
                callback()

    wrapper.instances = instances

    return wrapper


class Gadget:
    r"""
    Base class for creating gadgets.

    Parameters
    ----------
    size : Size, default: Size(10, 10)
        Size of gadget.
    pos : Point, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: False
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : SizeHint
        Size as a proportion of parent's height and width.
    pos_hint : PosHint
        Position as a proportion of parent's height and width.
    parent : Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App
        The running app.

    Methods
    -------
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(\*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    __bindings: dict[int, str] = {}
    """UID to property name mapping."""

    def __init__(
        self,
        *,
        size: Size = Size(10, 10),
        pos: Point = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = False,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        self.parent: Gadget | None = None
        """The gadget's parent."""
        self.children: _GadgetList = _GadgetList()
        """The gadget's children."""

        h, w = size
        self._size = Size(clamp(h, 0, None), clamp(w, 0, None))
        """Size of gadget."""
        self._pos = Point(*pos)
        """Position relative to parent."""
        self._size_hint: SizeHint = _normalize_size_hint(size_hint or {})
        """Gadget's size as a proportion of its parent's size."""
        self._pos_hint: PosHint = _normalize_pos_hint(pos_hint or {})
        """Gadget's position as a proportion of its parent's size."""
        self._is_transparent = is_transparent
        """Whether gadget is transparent."""
        self._is_visible = is_visible
        """Whether gadget is visible."""
        self._is_enabled = is_enabled
        """Whether gadget is enabled."""
        self._region_valid: bool = False
        """Whether current region is valid."""
        self._clipping_region: Region = Region()
        """Initial region clipped by ancestor regions."""
        self._root_region_before: Region = Region()
        """The root's region before gadget's region is removed from it."""
        self._region: Region = Region()
        """The visible portion of the gadget on the screen."""

    def __repr__(self):
        return (
            f"{type(self).__name__}(size={self.size}, pos={self.pos}, "
            f"size_hint={self._size_hint}, pos_hint={self._pos_hint}, "
            f"is_transparent={self.is_transparent}, is_visible={self.is_visible}, "
            f"is_enabled={self.is_enabled})"
        )

    @property
    def size(self) -> Size:
        """Size of gadget."""
        return self._size

    @size.setter
    @bindable
    def size(self, size: Size):
        if size == self._size:
            self._apply_pos_hints()
            return

        h, w = size
        size = Size(clamp(int(h), 0, None), clamp(int(w), 0, None))
        if self.root is None:
            self._size = size
        else:
            with self.root._render_lock:
                self._size = size
                self._invalidate_region()

        self._apply_pos_hints()
        for child in self.children:
            child.apply_hints()
        self.on_size()

    @property
    def height(self) -> int:
        """Height of gadget."""
        return self._size[0]

    @height.setter
    def height(self, height: int):
        self.size = height, self.width

    rows = height
    """Alias for :attr:`height`."""

    @property
    def width(self) -> int:
        """Width of gadget."""
        return self._size[1]

    @width.setter
    def width(self, width: int):
        self.size = self.height, width

    columns = width
    """Alias for :attr:`width`."""

    @property
    def pos(self) -> Point:
        """Position relative to parent."""
        return self._pos

    @pos.setter
    @bindable
    def pos(self, pos: Point):
        y, x = pos
        pos = Point(int(y), int(x))

        if self.root is None:
            self._pos = pos
        else:
            with self.root._render_lock:
                self._pos = pos
                self._invalidate_region()

    @property
    def top(self) -> int:
        """y-coordinate of top of gadget."""
        return self._pos[0]

    @top.setter
    def top(self, top: int):
        self.pos = top, self.left

    y = top
    """y-coordinate of top of gadget."""

    @property
    def left(self) -> int:
        """x-coordinate of left side of gadget."""
        return self._pos[1]

    @left.setter
    def left(self, left: int):
        self.pos = self.top, left

    x = left
    """x-coordinate of left side of gadget."""

    @property
    def bottom(self) -> int:
        """y-coordinate of bottom of gadget."""
        return self.top + self.height

    @bottom.setter
    def bottom(self, bottom: int):
        self.top = bottom - self.height

    @property
    def right(self) -> int:
        """x-coordinate of right side of gadget."""
        return self.left + self.width

    @right.setter
    def right(self, right: int):
        self.left = right - self.width

    @property
    def center(self) -> Point:
        """Position of center of gadget."""
        return self.pos + self.size.center

    @center.setter
    def center(self, center: Point):
        cy, cx = center
        h, w = self.size
        self.pos = Point(cy - h // 2, cx - w // 2)

    @property
    def absolute_pos(self) -> Point:
        """Absolute position on screen."""
        if self.parent is None:
            return self.pos
        return self.pos + self.parent.absolute_pos

    @property
    def size_hint(self) -> SizeHint:
        """Gadget's size as a proportion of its parent's size."""
        return MappingProxyType(self._size_hint)

    @size_hint.setter
    def size_hint(self, size_hint: SizeHint):
        self._size_hint = _normalize_size_hint(size_hint)
        self.apply_hints()

    @property
    def pos_hint(self) -> PosHint:
        """Gadget's position as a proportion of its parent's size."""
        return MappingProxyType(self._pos_hint)

    @pos_hint.setter
    def pos_hint(self, pos_hint: PosHint):
        self._pos_hint = _normalize_pos_hint(pos_hint)
        self.apply_hints()

    @property
    def is_transparent(self) -> bool:
        """Whether gadget is transparent."""
        return self._is_transparent

    @is_transparent.setter
    @bindable
    def is_transparent(self, is_transparent: bool):
        if is_transparent != self._is_transparent:
            self._is_transparent = is_transparent
            if self.root is not None:
                self._invalidate_region()
            self.on_transparency()

    @property
    def is_visible(self) -> bool:
        """
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
        """
        return self._is_visible

    @is_visible.setter
    @bindable
    def is_visible(self, is_visible: bool):
        if is_visible != self._is_visible:
            self._is_visible = is_visible
            if self.root is not None:
                self._invalidate_region()

    @property
    def is_enabled(self) -> bool:
        """
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.
        """
        return self._is_enabled

    @is_enabled.setter
    @bindable
    def is_enabled(self, is_enabled: bool):
        if is_enabled != self._is_enabled:
            self._is_enabled = is_enabled
            if self.root is not None:
                self._invalidate_region()

    @property
    def root(self) -> Self | None:
        """Return the root gadget if connected to gadget tree."""
        return self.parent and self.parent.root

    @property
    def app(self):
        """The running app."""
        return self.root.app

    def _render(self, canvas: NDArray[Cell]) -> None:
        """Render visible region of gadget."""

    def dispatch_key(self, key_event: KeyEvent) -> bool | None:
        """
        Dispatch a key press event until handled.

        A key press event is handled if a handler returns ``True``.

        Parameters
        ----------
        key_event : KeyEvent
            The key event to dispatch.

        Returns
        -------
        bool | None
            Whether the dispatch was handled.
        """
        return any(
            gadget.dispatch_key(key_event)
            for gadget in reversed(self.children)
            if gadget.is_enabled
        ) or self.on_key(key_event)

    def dispatch_mouse(self, mouse_event: MouseEvent) -> bool | None:
        """
        Dispatch a mouse event until handled.

        A mouse event is handled if a handler returns ``True``.

        Parameters
        ----------
        mouse_event : MouseEvent
            The mouse event to dispatch.

        Returns
        -------
        bool | None
            Whether the dispatch was handled.
        """
        return any(
            gadget.dispatch_mouse(mouse_event)
            for gadget in reversed(self.children)
            if gadget.is_enabled
        ) or self.on_mouse(mouse_event)

    def dispatch_paste(self, paste_event: PasteEvent) -> bool | None:
        """
        Dispatch a paste event until handled.

        A paste event is handled if a handler returns ``True``.

        Parameters
        ----------
        paste_event : PasteEvent
            The paste event to dispatch.

        Returns
        -------
        bool | None
            Whether the dispatch was handled.
        """
        return any(
            gadget.dispatch_paste(paste_event)
            for gadget in reversed(self.children)
            if gadget.is_enabled
        ) or self.on_paste(paste_event)

    def dispatch_terminal_focus(self, focus_event: FocusEvent) -> bool | None:
        """
        Dispatch a focus event until handled.

        A focus event is handled if a handler returns ``True``.

        Parameters
        ----------
        focus_event : FocusEvent
            The focus event to dispatch.

        Returns
        -------
        bool | None
            Whether the dispatch was handled.
        """
        return any(
            gadget.dispatch_terminal_focus(focus_event)
            for gadget in reversed(self.children)
            if gadget.is_enabled
        ) or self.on_terminal_focus(focus_event)

    def _invalidate_region(self, notify_root: bool = True) -> None:
        """Invalidate region and all children's regions."""
        self._region_valid = False
        if notify_root and self.root is not None:
            self.root._all_regions_valid = False
        for child in self.children:
            child._invalidate_region(False)

    def apply_hints(self) -> None:
        """
        Apply size and pos hints.

        This is called automatically when the gadget is added to the gadget tree and
        when the gadget's parent's size changes.
        """
        if self.parent is None:
            return

        parent_height, parent_width = self.parent.size
        height, width = self.size
        if self._size_hint["height_hint"] is not None:
            height = clamp(
                round_down(parent_height * self._size_hint["height_hint"])
                + self._size_hint["height_offset"],
                self._size_hint["min_height"],
                self._size_hint["max_height"],
            )

        if self._size_hint["width_hint"] is not None:
            width = clamp(
                round_down(parent_width * self._size_hint["width_hint"])
                + self._size_hint["width_offset"],
                self._size_hint["min_width"],
                self._size_hint["max_width"],
            )

        self.size = height, width  # `size` setter will call `_apply_pos_hints()`.

    def _apply_pos_hints(self) -> None:
        if self.parent is None:
            return

        parent_height, parent_width = self.parent.size
        height, width = self.size
        y, x = self.pos
        y_anchor, x_anchor = self._pos_hint["anchor"]

        if self._pos_hint["y_hint"] is not None:
            y = (
                round_down(parent_height * self._pos_hint["y_hint"])
                - round_down(height * y_anchor)
                + self._pos_hint["y_offset"]
            )

        if self._pos_hint["x_hint"] is not None:
            x = (
                round_down(parent_width * self._pos_hint["x_hint"])
                - round_down(width * x_anchor)
                + self._pos_hint["x_offset"]
            )
        self.pos = y, x

    def to_local(self, point: Point) -> Point:
        """
        Convert point in absolute coordinates to local coordinates.

        Parameters
        ----------
        point : Point
            Point in absolute (screen) coordinates.

        Returns
        -------
        Point
            The point in local coordinates.
        """
        return point - self.absolute_pos

    def collides_point(self, point: Point) -> bool:
        """
        Return true if point collides with visible portion of gadget.

        Parameters
        ----------
        point : Point
            A point.

        Returns
        -------
        bool
            Whether point collides with gadget.
        """
        if self.root is None or not (self.is_visible and self.is_enabled):
            return False

        return point in self._region or any(
            point in child._region
            for child in self.walk()
            if child.is_visible and child.is_enabled
        )

    def collides_gadget(self, other: Self) -> bool:
        """
        Return true if other is within gadget's bounding box.

        Parameters
        ----------
        other : Gadget
            Another gadget.

        Returns
        -------
        bool
            Whether other collides with gadget.
        """
        self_top, self_left = self.absolute_pos
        self_bottom = self_top + self.height
        self_right = self.left + self.width

        other_top, other_left = other.absolute_pos
        other_bottom = other_top + other.height
        other_right = other_left + other.width

        return not (
            self_top >= other_bottom
            or other_top >= self_bottom
            or self_left >= other_right
            or other_left >= self_right
        )

    def pull_to_front(self) -> None:
        """Move gadget to end of gadget stack so that it is drawn last."""
        if self.parent is not None:
            self.parent.children.remove(self)
            self.parent.children.append(self)

    def walk(self) -> Iterator[Self]:
        """
        Yield all descendents of this gadget (preorder traversal).

        Yields
        ------
        Gadget
            A descendent of this gadget.
        """
        for child in self.children:
            yield child
            yield from child.walk()

    def walk_reverse(self) -> Iterator[Self]:
        """
        Yield all descendents of this gadget (reverse postorder traversal).

        Yields
        ------
        Gadget
            A descendent of this gadget.
        """
        for child in reversed(self.children):
            yield from child.walk_reverse()
            yield child

    def ancestors(self) -> Iterator[Self]:
        """
        Yield all ancestors of this gadget.

        Yields
        ------
        Gadget
            An ancestor of this gadget.
        """
        if self.parent:
            yield self.parent
            yield from self.parent.ancestors()

    def add_gadget(self, gadget: Self) -> None:
        """
        Add a child gadget.

        Parameters
        ----------
        gadget : Gadget
            A gadget to add as a child.
        """
        self.children.append(gadget)
        gadget.parent = self

        if self.root is not None:
            gadget.on_add()

    def add_gadgets(self, *gadgets: Self) -> None:
        r"""
        Add multiple child gadgets.

        Parameters
        ----------
        \*gadgets : Gadget
            Gadgets to add as children. Can also accept a single iterable of gadgets.
        """
        if len(gadgets) == 1 and not isinstance(gadgets[0], Gadget):
            # Assume item is an iterable of gadgets.
            gadgets = gadgets[0]

        for gadget in gadgets:
            self.add_gadget(gadget)

    def remove_gadget(self, gadget: Self) -> None:
        """
        Remove a child gadget.

        Parameters
        ----------
        gadget : Gadget
            The gadget to remove from children.
        """
        if self.root is not None:
            gadget.on_remove()

        self.children.remove(gadget)
        gadget.parent = None

    def prolicide(self) -> None:
        """Recursively remove all children."""
        for child in self.children.copy():
            child.destroy()

    def destroy(self) -> None:
        """Remove this gadget and recursively remove all its children."""
        self.prolicide()
        if self.parent:
            self.parent.remove_gadget(self)

    def bind(self, prop: str, callback: Callable[[], None]) -> int:
        """
        Bind `callback` to a gadget property. When the property is updated, `callback`
        is called with no arguments.

        Parameters
        ----------
        prop : str
            The name of the gadget property.
        callback : Callable[[], None]
            Callback to bind to property.

        Returns
        -------
        int
            A unique id used to unbind the callback.
        """
        uid = next(_UID)
        setter = getattr(type(self), prop).fset
        bindings = setter.instances.setdefault(self, {})
        bindings[uid] = callback
        self.__bindings[uid] = prop
        return uid

    def unbind(self, uid: int) -> None:
        """
        Unbind a callback from a gadget property.

        Parameters
        ----------
        uid : int
            Unique id returned by the :meth:`bind` method.
        """
        prop = self.__bindings.pop(uid, None)
        if prop is None:
            return
        setter = getattr(type(self), prop).fset
        if self in setter.instances:
            setter.instances[self].pop(uid, None)

    @staticmethod
    def _tween_lerp(start, end, p) -> Real | Sequence[Real] | PosHint | SizeHint | None:
        """Lerp for none, sequences, and hints."""
        if start is None or end is None:
            return None

        if isinstance(start, Real) and isinstance(end, Real):
            value = lerp(start, end, p)
            if isinstance(start, int):
                return round_down(value)
            return value

        if isinstance(start, Sequence):
            return [
                Gadget._tween_lerp(start_value, end_value, p)
                for start_value, end_value in zip(start, end)
            ]

        if isinstance(start, MappingProxyType):
            return {
                key: Gadget._tween_lerp(start_value, end.get(key), p)
                for key, start_value in start.items()
            }

    async def tween(
        self,
        *,
        duration: float = 1.0,
        easing: Easing = "linear",
        on_start: Callable[[], None] | None = None,
        on_progress: Callable[[float], None] | None = None,
        on_complete: Callable[[], None] | None = None,
        **properties: Real | NDArray[np.number] | Sequence[Real] | PosHint | SizeHint,
    ) -> None:
        """
        Coroutine that sequentially updates gadget properties over a duration (in
        seconds).

        Parameters
        ----------
        duration : float, default: 1.0
            The duration of the tween in seconds.
        easing : Easing, default: "linear"
            The easing used for tweening.
        on_start : Callable[[], None] | None, default: None
            Called when tween starts.
        on_progress : Callable[[float], None] | None, default: None
            Called as tween updates with current progress.
        on_complete : Callable[[], None] | None, default: None
            Called when tween completes.
        **properties : Real | NDArray[np.number] | Sequence[Real] | PosHint | SizeHint
            Gadget properties' target values. E.g., to smoothly tween a gadget's
            position to (5, 10) over 2.5 seconds, specify the `pos` property as a
            keyword-argument:
            ``await gadget.tween(pos=(5, 10), duration=2.5, easing="out_bounce")``

        Notes
        -----
        Tweened values will be coerced to match the type of the initial value of their
        corresponding property.

        Non-numeric values will be set immediately.

        Warnings
        --------
        Running several tweens on the same properties concurrently will probably result
        in unexpected behavior.
        """
        end_time = perf_counter() + duration
        easing_function = EASINGS[easing]
        start_values = tuple(getattr(self, attr) for attr in properties)

        if pos_hint := properties.get("pos_hint"):
            properties["pos_hint"] = _normalize_pos_hint(pos_hint)
        if size_hint := properties.get("size_hint"):
            properties["size_hint"] = _normalize_size_hint(size_hint)
        if on_start is not None:
            on_start()

        while (current_time := perf_counter()) < end_time:
            p = easing_function(1 - (end_time - current_time) / duration)

            for start_value, (prop, target) in zip(start_values, properties.items()):
                setattr(self, prop, Gadget._tween_lerp(start_value, target, p))

            if on_progress is not None:
                on_progress(p)

            await asyncio.sleep(0)

        for prop, target in properties.items():
            setattr(self, prop, target)

        if on_complete is not None:
            on_complete()

    def on_size(self) -> None:
        """Update gadget after a resize."""

    def on_transparency(self) -> None:
        """Update gadget after transparency is enabled/disabled."""

    def on_add(self) -> None:
        """Update gadget after being added to the gadget tree."""
        self.apply_hints()
        for child in self.children:
            child.on_add()

    def on_remove(self) -> None:
        """Update gadget after being removed from the gadget tree."""
        for child in self.children:
            child.on_remove()

    def on_key(self, key_event: KeyEvent) -> bool | None:
        """
        Handle a key press event.

        Handled key presses should return ``True``.

        Parameters
        ----------
        key_event : KeyEvent
            The key event to handle.

        Returns
        -------
        bool | None
            Whether the key event was handled.
        """

    def on_mouse(self, mouse_event: MouseEvent) -> bool | None:
        """
        Handle a mouse event.

        Handled mouse events should return ``True``.

        Parameters
        ----------
        mouse_event : MouseEvent
            The mouse event to handle.

        Returns
        -------
        bool | None
            Whether the mouse event was handled.
        """

    def on_paste(self, paste_event: PasteEvent) -> bool | None:
        """
        Handle a paste event.

        Handled paste events should return ``True``.

        Parameters
        ----------
        paste_event : PasteEvent
            The paste event to handle.

        Returns
        -------
        bool | None
            Whether the paste event was handled.
        """

    def on_terminal_focus(self, focus_event: FocusEvent) -> bool | None:
        """
        Handle a focus event.

        Handled focus events should return ``True``.

        Parameters
        ----------
        focus_event : FocusEvent
            The focus event to handle.

        Returns
        -------
        bool | None
            Whether the focus event was handled.
        """



"""A color picker gadget."""

from collections.abc import Callable
from itertools import pairwise

import numpy as np

from ..colors import (
    ABLACK,
    ABLUE,
    ACYAN,
    AGREEN,
    AMAGENTA,
    ARED,
    AWHITE,
    AYELLOW,
    RED,
    WHITE,
    AColor,
    Color,
    gradient,
)
from .behaviors.grabbable import Grabbable
from .behaviors.themable import Themable
from .button import Button
from .gadget import Gadget, Point, PosHint, Size, SizeHint
from .graphics import Graphics
from .pane import Pane
from .text import Text, new_cell

__all__ = ["ColorPicker", "Point", "Size"]

GRAD = tuple(pairwise([ARED, AYELLOW, AGREEN, ACYAN, ABLUE, AMAGENTA, ARED]))


class _ShadeSelector(Grabbable, Graphics):
    def __init__(self, color_swatch, label, **kwargs):
        super().__init__(**kwargs)

        self._shade_indicator = Text(size=(1, 1), is_transparent=True, default_cell="○")
        self._shade_hint = 0.0, 1.0
        self.add_gadget(self._shade_indicator)

        self.color_swatch = color_swatch
        self.label = label
        self.update_hue(ARED)

    def on_size(self):
        h, w = self._size
        hh, wh = self._shade_hint

        self.texture = np.zeros((h * 2, w, 4), dtype=np.uint8)
        self._shade_indicator.pos = round((h - 1) * hh), round((w - 1) * wh)

        self.update_hue(self.hue)

    def update_hue(self, hue: AColor):
        self.hue = hue

        h, w = self._size
        if w == 0:
            return
        left_side = gradient(AWHITE, ABLACK, 2 * h)
        right_side = gradient(hue, ABLACK, 2 * h)

        for row, left, right in zip(self.texture, left_side, right_side):
            row[:] = gradient(left, right, w)

        self.update_swatch_label()

    def update_swatch_label(self):
        y, x = self._shade_indicator.pos

        r, g, b = self.texture[y * 2, x, :3].tolist()

        self.color_swatch.bg_color = r, g, b

        self.label.add_str(hex(r * 2**16 + g * 2**8 + b)[2:], pos=(1, 1))
        self.label.add_str(f"R: {r:>3}", pos=(3, 1))
        self.label.add_str(f"G: {g:>3}", pos=(4, 1))
        self.label.add_str(f"B: {b:>3}", pos=(5, 1))

    def grab(self, mouse_event):
        super().grab(mouse_event)
        self.grab_update(mouse_event)

    def grab_update(self, mouse_event):
        if self.collides_point(mouse_event.pos):
            y, x = self._shade_indicator.pos = self.to_local(mouse_event.pos)
            h, w = self.size
            self._shade_hint = (
                0 if h <= 1 else y / (h - 1),
                0 if w <= 1 else x / (w - 1),
            )
            self.update_swatch_label()


class _HueSelector(Grabbable, Graphics):
    def __init__(self, shade_selector, **kwargs):
        super().__init__(**kwargs)
        self.shade_selector = shade_selector

        self._hue_hint = 0.0
        self._hue_indicator = Text(
            size=(1, 1), default_cell=new_cell(fg_color=WHITE, bg_color=RED)
        )
        self._hue_indicator.add_str("▼")

        self.add_gadget(self._hue_indicator)

    def on_size(self):
        h, w = self._size

        self.texture = np.zeros((h * 2, w, 4), dtype=np.uint8)

        d, r = divmod(w, 6)

        rainbow = []
        for i, (a, b) in enumerate(GRAD):
            rainbow.extend(gradient(a, b, d + (i < r)))

        self.texture[:] = rainbow

        self._hue_indicator.x = round(self._hue_hint * w)
        self.update_hue()

    def update_hue(self):
        x = self._hue_indicator.x
        self._hue_indicator.canvas["bg_color"] = self.texture[0, x, :3]
        self.shade_selector.update_hue(AColor(*self.texture[0, x]))

    def grab(self, mouse_event):
        super().grab(mouse_event)
        self.grab_update(mouse_event)

    def grab_update(self, mouse_event):
        if self.collides_point(mouse_event.pos):
            x = self._hue_indicator.x = self.to_local(mouse_event.pos).x
            self._hue_hint = 0 if self.width <= 1 else x / (self.width - 1)
            self.update_hue()


class ColorPicker(Themable, Gadget):
    r"""
    A color picker gadget.

    Parameters
    ----------
    ok_callback : Callable[[Color], None], default: lambda color: None
        Called with currently selected color when "OK" button is released.
    alpha : float, default: 1.0
        Transparency of gadget.
    size : Size, default: Size(10, 10)
        Size of gadget.
    pos : Point, default: Point(0, 0)
        Position of upper-left corner in parent.
    size_hint : SizeHint | None, default: None
        Size as a proportion of parent's height and width.
    pos_hint : PosHint | None, default: None
        Position as a proportion of parent's height and width.
    is_transparent : bool, default: False
        Whether gadget is transparent.
    is_visible : bool, default: True
        Whether gadget is visible. Gadget will still receive input events if not
        visible.
    is_enabled : bool, default: True
        Whether gadget is enabled. A disabled gadget is not painted and doesn't receive
        input events.

    Attributes
    ----------
    alpha : float
        Transparency of gadget.
    size : Size
        Size of gadget.
    height : int
        Height of gadget.
    rows : int
        Alias for :attr:`height`.
    width : int
        Width of gadget.
    columns : int
        Alias for :attr:`width`.
    pos : Point
        Position of upper-left corner.
    top : int
        y-coordinate of top of gadget.
    y : int
        y-coordinate of top of gadget.
    left : int
        x-coordinate of left side of gadget.
    x : int
        x-coordinate of left side of gadget.
    bottom : int
        y-coordinate of bottom of gadget.
    right : int
        x-coordinate of right side of gadget.
    center : Point
        Position of center of gadget.
    absolute_pos : Point
        Absolute position on screen.
    size_hint : SizeHint
        Size as a proportion of parent's height and width.
    pos_hint : PosHint
        Position as a proportion of parent's height and width.
    parent: Gadget | None
        Parent gadget.
    children : list[Gadget]
        Children gadgets.
    is_transparent : bool
        Whether gadget is transparent.
    is_visible : bool
        Whether gadget is visible.
    is_enabled : bool
        Whether gadget is enabled.
    root : Gadget | None
        If gadget is in gadget tree, return the root gadget.
    app : App
        The running app.

    Methods
    -------
    update_theme()
        Paint the gadget with current theme.
    apply_hints()
        Apply size and pos hints.
    to_local(point)
        Convert point in absolute coordinates to local coordinates.
    collides_point(point)
        Return true if point collides with visible portion of gadget.
    collides_gadget(other)
        Return true if other is within gadget's bounding box.
    pull_to_front()
        Move to end of gadget stack so gadget is drawn last.
    walk()
        Yield all descendents of this gadget (preorder traversal).
    walk_reverse()
        Yield all descendents of this gadget (reverse postorder traversal).
    ancestors()
        Yield all ancestors of this gadget.
    add_gadget(gadget)
        Add a child gadget.
    add_gadgets(\*gadgets)
        Add multiple child gadgets.
    remove_gadget(gadget)
        Remove a child gadget.
    prolicide()
        Recursively remove all children.
    destroy()
        Remove this gadget and recursively remove all its children.
    bind(prop, callback)
        Bind `callback` to a gadget property.
    unbind(uid)
        Unbind a callback from a gadget property.
    tween(...)
        Sequentially update gadget properties over time.
    on_size()
        Update gadget after a resize.
    on_transparency()
        Update gadget after transparency is enabled/disabled.
    on_add()
        Update gadget after being added to the gadget tree.
    on_remove()
        Update gadget after being removed from the gadget tree.
    on_key(key_event)
        Handle a key press event.
    on_mouse(mouse_event)
        Handle a mouse event.
    on_paste(paste_event)
        Handle a paste event.
    on_terminal_focus(focus_event)
        Handle a focus event.
    """

    def __init__(
        self,
        *,
        ok_callback: Callable[[Color], None] = lambda color: None,
        alpha: float = 1.0,
        size: Size = Size(10, 10),
        pos: Point = Point(0, 0),
        size_hint: SizeHint | None = None,
        pos_hint: PosHint | None = None,
        is_transparent: bool = False,
        is_visible: bool = True,
        is_enabled: bool = True,
    ):
        self.color_swatch = Pane(pos=(1, 1), bg_color=RED, is_transparent=False)
        self.label = Text(size=(9, 8), alpha=0, is_transparent=True)
        self.shades = _ShadeSelector(
            pos=(1, 1),
            color_swatch=self.color_swatch,
            label=self.label,
            is_transparent=False,
        )
        self.hues = _HueSelector(
            pos=(1, 1), shade_selector=self.shades, is_transparent=False
        )
        ok_button = Button(
            label="OK",
            size=(1, 6),
            pos=(7, 1),
            callback=lambda: ok_callback(self.color_swatch.bg_color),
        )
        self._container = Pane(
            size_hint={"height_hint": 1.0, "width_hint": 1.0},
            is_transparent=is_transparent,
        )
        super().__init__(
            size=size,
            pos=pos,
            size_hint=size_hint,
            pos_hint=pos_hint,
            is_transparent=is_transparent,
            is_visible=is_visible,
            is_enabled=is_enabled,
        )
        self.alpha = alpha
        self.label.add_gadget(ok_button)
        self._container.add_gadgets(
            self.color_swatch, self.hues, self.shades, self.label
        )
        self.add_gadget(self._container)

    @property
    def alpha(self) -> float:
        """Transparency of gadget."""
        return self._container.alpha

    @alpha.setter
    def alpha(self, alpha: float):
        self._container.alpha = alpha

    def on_transparency(self) -> None:
        """Update gadget after transparency is enabled/disabled."""
        self._container.is_transparent = self.is_transparent

    def on_size(self):
        """Resize and reposition children."""
        h, w = self._size

        self.shades.size = max(10, h - 4), max(20, w - 11)

        self.color_swatch.size = max(2, h - 12), 8
        self.color_swatch.left = self.shades.right + 1

        self.hues.size = 1, self.shades.width
        self.hues.top = self.shades.bottom + 1

        self.label.top = self.color_swatch.bottom + 1
        self.label.left = self.shades.right + 1

    def update_theme(self):
        """Paint the gadget with current theme."""
        primary = self.color_theme.primary
        self._container.bg_color = primary.bg
        self.label.default_fg_color = self.label.canvas["fg_color"] = primary.fg
        self.label.default_bg_color = self.label.canvas["bg_color"] = primary.bg



"""A windows VT100 terminal."""

import asyncio
import re
from collections.abc import Callable
from ctypes import Structure, Union, byref, windll
from ctypes.wintypes import BOOL, CHAR, DWORD, HANDLE, LONG, LPVOID, SHORT, WCHAR
from typing import Final

from ..geometry import Size
from .events import Event, ResizeEvent
from .vt100_terminal import Vt100Terminal

CTRL_SPACE_RE: Final = re.compile(r"\x00+")
CTRL_ALT_SPACE_RE: Final = re.compile(r"\x00*\x1b\x00+")
STDIN = windll.kernel32.GetStdHandle(DWORD(-10))
STDOUT = windll.kernel32.GetStdHandle(DWORD(-11))
KEY_EVENT: Final = 1
WINDOW_BUFFER_SIZE_EVENT: Final = 4
# See: https://msdn.microsoft.com/pl-pl/library/windows/desktop/ms686033(v=vs.85).aspx
ENABLE_VIRTUAL_TERMINAL_PROCESSING = 0x0004
ENABLE_VIRTUAL_TERMINAL_INPUT = 0x0200


class UNICODE_OR_ASCII(Union):
    """Windows character."""

    _fields_ = [("AsciiChar", CHAR), ("UnicodeChar", WCHAR)]


class KEY_EVENT_RECORD(Structure):
    """
    Windows key event record.

    References
    ----------
    http://msdn.microsoft.com/en-us/library/windows/desktop/ms684166(v=vs.85).aspx
    """

    _fields_ = [
        ("KeyDown", LONG),
        ("RepeatCount", SHORT),
        ("VirtualKeyCode", SHORT),
        ("VirtualScanCode", SHORT),
        ("uChar", UNICODE_OR_ASCII),
        ("ControlKeyState", LONG),
    ]


class COORD(Structure):
    """
    Windows coord struct.

    References
    ----------
    http://msdn.microsoft.com/en-us/library/windows/desktop/ms682119(v=vs.85).aspx
    """

    _fields_ = [("X", SHORT), ("Y", SHORT)]


class MOUSE_EVENT_RECORD(Structure):
    """
    Windows mouse event record.

    References
    ----------
    http://msdn.microsoft.com/en-us/library/windows/desktop/ms684239(v=vs.85).aspx
    """

    _fields_ = [
        ("MousePosition", COORD),
        ("ButtonState", LONG),
        ("ControlKeyState", LONG),
        ("EventFlags", LONG),
    ]


class WINDOW_BUFFER_SIZE_RECORD(Structure):
    """
    Windows buffer size record.

    References
    ----------
    http://msdn.microsoft.com/en-us/library/windows/desktop/ms687093(v=vs.85).aspx
    """

    _fields_ = [("Size", COORD)]


class MENU_EVENT_RECORD(Structure):
    """
    Windows menu event record.

    References
    ----------
    http://msdn.microsoft.com/en-us/library/windows/desktop/ms684213(v=vs.85).aspx
    """

    _fields_ = [("CommandId", LONG)]


class FOCUS_EVENT_RECORD(Structure):
    """
    Windows focus event record.

    References
    ----------
    http://msdn.microsoft.com/en-us/library/windows/desktop/ms683149(v=vs.85).aspx
    """

    _fields_ = [("SetFocus", LONG)]


class EVENT_RECORD(Union):
    """Windows event record."""

    _fields_ = [
        ("KeyEvent", KEY_EVENT_RECORD),
        ("MouseEvent", MOUSE_EVENT_RECORD),
        ("WindowBufferSizeEvent", WINDOW_BUFFER_SIZE_RECORD),
        ("MenuEvent", MENU_EVENT_RECORD),
        ("FocusEvent", FOCUS_EVENT_RECORD),
    ]


class INPUT_RECORD(Structure):
    """
    Windows input record.

    References
    ----------
    http://msdn.microsoft.com/en-us/library/windows/desktop/ms683499(v=vs.85).aspx
    """

    _fields_ = [("EventType", SHORT), ("Event", EVENT_RECORD)]


class SECURITY_ATTRIBUTES(Structure):
    """
    Windows security attributes.

    References
    ----------
    http://msdn.microsoft.com/en-us/library/windows/desktop/aa379560(v=vs.85).aspx
    """

    _fields_ = [
        ("nLength", DWORD),
        ("lpSecurityDescriptor", LPVOID),
        ("bInheritHandle", BOOL),
    ]


class WindowsTerminal(Vt100Terminal):
    """
    A windows VT100 terminal.

    Attributes
    ----------
    in_alternate_screen : bool
        Whether the alternate screen buffer is enabled.

    Methods
    -------
    process_stdin()
        Read from stdin and feed data into input parser to generate events.
    raw_mode()
        Set terminal to raw mode.
    restore_console()
        Restore console to its original mode.
    attach(event_handler)
        Start generating events from stdin.
    unattach()
        Stop generating events from stdin.
    events()
        Return a list of input events and reset the event buffer.
    get_size()
        Get terminal size.
    flush()
        Write buffer to output stream and flush.
    set_title(title)
        Set terminal title.
    enter_alternate_screen()
        Enter alternate screen buffer.
    exit_alternate_screen()
        Exit alternate screen buffer.
    enable_mouse_support()
        Enable mouse support in terminal.
    disable_mouse_support()
        Disable mouse support in terminal.
    reset_attributes()
        Reset character attributes.
    enable_bracketed_paste()
        Enable bracketed paste in terminal.
    disable_bracketed_paste()
        Disable bracketed paste in terminal.
    show_cursor()
        Show cursor in terminal.
    hide_cursor()
        Hide cursor in terminal.
    enable_reporting_focus()
        Enable reporting terminal focus.
    disable_reporting_focus()
        Disable reporting terminal focus.
    request_cursor_position_report()
        Report current cursor position.
    request_foreground_color()
        Report terminal foreground color.
    request_background_color()
        Report terminal background color.
    request_device_attributes()
        Report device attributes.
    expect_dsr()
        Return whether a device status report is expected.
    move_cursor(pos)
        Move cursor to ``pos``.
    erase_in_display(n)
        Clear part of the screen.
    """

    def _feed(self, data: str) -> None:
        # Some versions of Windows Terminal generate spurious null characters for a few
        # input events. For instance, ctrl+" " generates 3 null characters instead of 1
        # and paste events generate a null character before each "shifted" character.
        # For most inputs, null characters can just be ignored.

        # ! The first two conditions assume the sequences for these events appear fully
        # ! and alone in stdin. Most of the time, key events occur slowly enough for
        # ! this to be true. Failing on other cases is acceptable here to keep this
        # ! logic simple.
        if CTRL_SPACE_RE.fullmatch(data):
            super()._feed("\x00")
        elif CTRL_ALT_SPACE_RE.fullmatch(data):
            super()._feed("\x1b\x00")
        else:
            super()._feed(data.replace("\x00", ""))

    def process_stdin(self) -> None:
        """Read from stdin and feed data into input parser to generate events."""
        nevents = DWORD()
        windll.kernel32.GetNumberOfConsoleInputEvents(STDIN, byref(nevents))
        InputRecordArray = INPUT_RECORD * nevents.value
        input_records = InputRecordArray()
        windll.kernel32.ReadConsoleInputW(
            STDIN, input_records, nevents.value, byref(DWORD())
        )
        chars = []
        for input_record in input_records:
            if input_record.EventType == KEY_EVENT:
                key_event = input_record.Event.KeyEvent
                if not key_event.KeyDown:
                    continue
                if key_event.ControlKeyState and not key_event.VirtualKeyCode:
                    continue
                chars.append(key_event.uChar.UnicodeChar)
            elif input_record.EventType == WINDOW_BUFFER_SIZE_EVENT:
                self._purge(chars)
                size = input_record.Event.WindowBufferSizeEvent.Size
                self._event_buffer.append(ResizeEvent(Size(size.Y, size.X)))
        self._purge(chars)

    def _purge(self, chars: list[str]):
        data = (
            "".join(chars).encode("utf-16", "surrogatepass").decode("utf-16")
        )  # Merge surrogate pairs.
        chars.clear()
        self._feed(data)

    def raw_mode(self) -> None:
        """Set terminal to raw mode."""
        self._original_output_mode = DWORD()
        windll.kernel32.GetConsoleMode(STDOUT, byref(self._original_output_mode))
        windll.kernel32.SetConsoleMode(
            STDOUT,
            self._original_output_mode.value | ENABLE_VIRTUAL_TERMINAL_PROCESSING,
        )
        self._original_input_mode = DWORD()
        windll.kernel32.GetConsoleMode(STDIN, byref(self._original_input_mode))
        windll.kernel32.SetConsoleMode(STDIN, ENABLE_VIRTUAL_TERMINAL_INPUT)

    def restore_console(self) -> None:
        """Restore console to its original mode."""
        windll.kernel32.SetConsoleMode(STDIN, self._original_input_mode)
        windll.kernel32.SetConsoleMode(STDOUT, self._original_output_mode)
        del self._original_input_mode, self._original_output_mode

    def attach(self, event_handler: Callable[[list[Event]], None]) -> None:
        """
        Start generating events from stdin.

        Parameters
        ----------
        event_handler : Callable[[list[Event]], None]
            Callable that handles input events.
        """
        self._event_buffer.clear()
        self._event_handler = event_handler
        loop = asyncio.get_running_loop()
        wait_for = windll.kernel32.WaitForMultipleObjects
        self._remove_event = HANDLE(
            windll.kernel32.CreateEventA(
                SECURITY_ATTRIBUTES(), BOOL(True), BOOL(False), None
            )
        )
        EVENTS = (HANDLE * 2)(self._remove_event, STDIN)

        def ready():
            try:
                self.process_stdin()
                event_handler(self.events())
            finally:
                loop.run_in_executor(None, wait)

        def wait():
            if wait_for(2, EVENTS, BOOL(False), DWORD(-1)) == 0:
                windll.kernel32.CloseHandle(self._remove_event)
                del self._remove_event
            else:
                loop.call_soon_threadsafe(ready)

        loop.run_in_executor(None, wait)

    def unattach(self) -> None:
        """Stop generating events from stdin."""
        self._event_handler = None
        windll.kernel32.SetEvent(self._remove_event)


def is_vt100_enabled() -> bool:
    """Return whether VT100 escape sequences are supported."""
    original_mode = DWORD()
    windll.kernel32.GetConsoleMode(STDOUT, byref(original_mode))

    try:
        return bool(
            windll.kernel32.SetConsoleMode(STDOUT, ENABLE_VIRTUAL_TERMINAL_PROCESSING)
        )
    finally:
        windll.kernel32.SetConsoleMode(STDOUT, original_mode)
